@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2003, 2004, 2005, 2006, 2007, 2008
+ * Copyright (C) 2003, 2004, 2005, 2006, 2007, 2008, 2009
  * Robert Lougher <rob@lougher.org.uk>.
  *
  * This file is part of JamVM.
@@ -182,13 +182,6 @@ static Thread *finalizer_thread;
 /* Pre-allocated OutOfMemoryError */
 static Object *oom;
 
-/* Cached primitive type array classes -- used to speed up
-   primitive array allocation */
-static Class *bool_array_class = NULL, *byte_array_class = NULL;
-static Class *char_array_class = NULL, *short_array_class = NULL;
-static Class *int_array_class = NULL, *float_array_class = NULL;
-static Class *double_array_class = NULL, *long_array_class = NULL;
-
 /* Field offsets and method table indexes used for finalization and
    reference handling. Cached in class.c */
 extern int ref_referent_offset;
@@ -226,7 +219,9 @@ static int sys_page_size;
 
 #define LIST_INCREMENT          100
 
-#define LOG_BYTESPERMARK        LOG_OBJECT_GRAIN /* 1 mark entry for every OBJECT_GRAIN bytes of heap */
+/* 1 mark entry for every OBJECT_GRAIN bytes of heap */
+#define LOG_BYTESPERMARK        LOG_OBJECT_GRAIN
+
 #define BITSPERMARK             2
 #define LOG_BITSPERMARK         1
 #define LOG_MARKSIZEBITS        5
@@ -234,15 +229,22 @@ static int sys_page_size;
 
 /* Macros for manipulating the mark bit array */
 
-#define MARKENTRY(ptr)     ((((char*)ptr)-heapbase)>>(LOG_BYTESPERMARK+LOG_MARKSIZEBITS-LOG_BITSPERMARK))
+#define MARKENTRY(ptr)     ((((char*)ptr)-heapbase)>> \
+                           (LOG_BYTESPERMARK+LOG_MARKSIZEBITS-LOG_BITSPERMARK))
+
 #define MARKOFFSET(ptr)    ((((((char*)ptr)-heapbase)>>LOG_BYTESPERMARK)& \
-                                                ((MARKSIZEBITS>>LOG_BITSPERMARK)-1))<<LOG_BITSPERMARK)
+                           ((MARKSIZEBITS>>LOG_BITSPERMARK)-1)) \
+                            <<LOG_BITSPERMARK)
+
 #define MARK(ptr,mark)     markBits[MARKENTRY(ptr)]|=mark<<MARKOFFSET(ptr);
 
-#define SET_MARK(ptr,mark) markBits[MARKENTRY(ptr)]=(markBits[MARKENTRY(ptr)]& \
-                                                ~(((1<<BITSPERMARK)-1)<<MARKOFFSET(ptr)))|mark<<MARKOFFSET(ptr)
+#define SET_MARK(ptr,mark) markBits[MARKENTRY(ptr)]= \
+                           (markBits[MARKENTRY(ptr)]& \
+                           ~(((1<<BITSPERMARK)-1)<<MARKOFFSET(ptr)))| \
+                           mark<<MARKOFFSET(ptr)
 
-#define IS_MARKED(ptr)     ((markBits[MARKENTRY(ptr)]>>MARKOFFSET(ptr))&((1<<BITSPERMARK)-1))
+#define IS_MARKED(ptr)     ((markBits[MARKENTRY(ptr)]>> \
+                           MARKOFFSET(ptr))&((1<<BITSPERMARK)-1))
 
 #define IS_HARD_MARKED(ptr)      (IS_MARKED(ptr) == HARD_MARK)
 #define IS_PHANTOM_MARKED(ptr)   (IS_MARKED(ptr) == PHANTOM_MARK)
@@ -251,7 +253,8 @@ static int sys_page_size;
                         (((char*)ptr) < heaplimit) && \
                         !(((uintptr_t)ptr)&(OBJECT_GRAIN-1))
 
-#define MIN_OBJECT_SIZE ((sizeof(Object)+HEADER_SIZE+OBJECT_GRAIN-1)&~(OBJECT_GRAIN-1))
+#define MIN_OBJECT_SIZE ((sizeof(Object)+HEADER_SIZE+OBJECT_GRAIN-1)& \
+                        ~(OBJECT_GRAIN-1))
 
 void allocMarkBits() {
     int no_of_bits = (heaplimit-heapbase)>>(LOG_BYTESPERMARK-LOG_BITSPERMARK);
@@ -271,7 +274,7 @@ void initialiseAlloc(InitArgs *args) {
 #ifdef USE_MALLOC
     /* Don't use mmap - malloc max heap size */
     char *mem = (char*)malloc(args->max_heap);
-    min = max;
+    args->min_heap = args->max_heap;
     if(mem == NULL) {
 #else
     char *mem = (char*)mmap(0, args->max_heap, PROT_READ|PROT_WRITE,
@@ -423,7 +426,7 @@ void markClassData(Class *class, int mark, int mark_soft_refs) {
         for(i = 0; i < cb->fields_count; i++, fb++)
             if((fb->access_flags & ACC_STATIC) &&
                         ((*fb->type == 'L') || (*fb->type == '['))) {
-                Object *ob = (Object *)fb->static_value;
+                Object *ob = fb->u.static_value.p;
                 TRACE_GC("Field %s %s object @%p\n", fb->name, fb->type, ob);
                 if(ob != NULL && mark > IS_MARKED(ob))
                     markChildren(ob, mark, mark_soft_refs);
@@ -455,7 +458,7 @@ void markChildren(Object *ob, int mark, int mark_soft_refs) {
 
     if(cb->name[0] == '[') {
         if((cb->name[1] == 'L') || (cb->name[1] == '[')) {
-            Object **body = ARRAY_DATA(ob);
+            Object **body = ARRAY_DATA(ob, Object*);
             int len = ARRAY_LEN(ob);
             int i;
             TRACE_GC("Scanning Array object @%p class is %s len is %d\n", ob, cb->name, len);
@@ -471,7 +474,6 @@ void markChildren(Object *ob, int mark, int mark_soft_refs) {
             TRACE_GC("Array object @%p class is %s  - Not Scanning...\n", ob, cb->name);
         }
     } else {
-        uintptr_t *body = INST_DATA(ob);
         int i;
 
         if(IS_CLASS_CLASS(cb)) {
@@ -487,7 +489,7 @@ void markChildren(Object *ob, int mark, int mark_soft_refs) {
                     markVMThrowable(ob, mark, mark_soft_refs);
                 } else
                     if(IS_REFERENCE(cb)) {
-                        Object *referent = (Object *)body[ref_referent_offset];
+                        Object *referent = INST_DATA(ob, Object*, ref_referent_offset);
 
                         TRACE_GC("Mark found Reference object @%p class %s flags %d referent %p\n",
                                  ob, cb->name, cb->flags, referent);
@@ -522,12 +524,12 @@ void markChildren(Object *ob, int mark, int mark_soft_refs) {
             int offset = cb->refs_offsets_table[i].start;
             int end = cb->refs_offsets_table[i].end;
 
-            for(; offset < end; offset++) {
-                Object *ob = (Object *)body[offset];
-                TRACE_GC("Offset %d reference @%p\n", offset, ob);
+            for(; offset < end; offset += sizeof(Object*)) {
+                Object *ref = INST_DATA(ob, Object*, offset);
+                TRACE_GC("Offset %d reference @%p\n", offset, ref);
 
-                if(ob != NULL && mark > IS_MARKED(ob))
-                    markChildren(ob, mark, mark_soft_refs);
+                if(ref != NULL && mark > IS_MARKED(ref))
+                    markChildren(ref, mark, mark_soft_refs);
             }
         }
     }
@@ -652,7 +654,7 @@ int handleMarkedSpecial(Object *ob) {
     int cleared = FALSE;
 
     if(IS_REFERENCE(cb)) {
-        Object *referent = (Object *)INST_DATA(ob)[ref_referent_offset];
+        Object *referent = INST_DATA(ob, Object*, ref_referent_offset);
 
         if(referent != NULL) {
             int ref_mark = IS_MARKED(referent);
@@ -668,14 +670,14 @@ int handleMarkedSpecial(Object *ob) {
                     goto out;
 
                 TRACE_GC("FREE: Clearing the referent field.\n");
-                INST_DATA(ob)[ref_referent_offset] = 0;
+                INST_DATA(ob, Object*, ref_referent_offset) = NULL;
                 cleared = TRUE;
             }
 
             /* If the reference has a queue, add it to the list for enqueuing
                by the Reference Handler thread. */
 
-            if((Object *)INST_DATA(ob)[ref_queue_offset] != NULL) {
+            if(INST_DATA(ob, Object*, ref_queue_offset) != NULL) {
                 TRACE_GC("FREE: Adding to list for enqueuing.\n");
 
                 ADD_TO_OBJECT_LIST(reference, ob);
@@ -705,7 +707,7 @@ void handleUnmarkedSpecial(Object *ob) {
                 /* Free the native thread structure (see comment
                    in detachThread (thread.c) */
                 TRACE_GC("FREE: Freeing native thread for VMThread object %p\n", ob);
-                gcPendingFree(threadSelf0(ob));
+                gcPendingFree(vmThread2Thread(ob));
             }
 }
 
@@ -925,13 +927,16 @@ void addConservativeRoots2Hash() {
     }
 }
 
-void registerStaticObjectRefLocked(Object **ref) {
+void registerStaticObjectRefLocked(Object **ref, Object *obj) {
     Thread *self = threadSelf();
 
     disableSuspend(self);
     lockVMLock(registered_refs_lock, self);
 
-    registerStaticObjectRef(ref);
+    if(*ref == NULL) {
+        *ref = obj;
+        registerStaticObjectRef(ref);
+    }
 
     unlockVMLock(registered_refs_lock, self);
     enableSuspend(self);
@@ -1023,17 +1028,17 @@ void threadClassData(Class *class, Class *new_addr) {
         threadReference(&cb->class_loader);
 
     if(cb->super != NULL)
-        threadReference((Object**)&cb->super);
+        threadReference(&cb->super);
 
     for(i = 0; i < cb->interfaces_count; i++)
         if(cb->interfaces[i] != NULL)
-            threadReference((Object**)&cb->interfaces[i]);
+            threadReference(&cb->interfaces[i]);
 
     if(IS_ARRAY(cb))
-        threadReference((Object**)&cb->element_class);
+        threadReference(&cb->element_class);
 
     for(i = 0; i < cb->imethod_table_size; i++)
-        threadReference((Object**)&cb->imethod_table[i].interface);
+        threadReference(&cb->imethod_table[i].interface);
 
     TRACE_COMPACT("Threading static fields for class %s\n", cb->name);
 
@@ -1043,7 +1048,7 @@ void threadClassData(Class *class, Class *new_addr) {
         for(i = 0; i < cb->fields_count; i++, fb++)
             if((fb->access_flags & ACC_STATIC) &&
                         ((*fb->type == 'L') || (*fb->type == '['))) {
-                Object **ob = (Object **)&fb->static_value;
+                Object **ob = (Object**)fb->u.static_value.data;
                 TRACE_COMPACT("Field %s %s object @%p\n", fb->name, fb->type, *ob);
                 if(*ob != NULL)
                     threadReference(ob);
@@ -1078,7 +1083,7 @@ int threadChildren(Object *ob, Object *new_addr) {
 
     if(cb->name[0] == '[') {
         if((cb->name[1] == 'L') || (cb->name[1] == '[')) {
-            Object **body = ARRAY_DATA(ob);
+            Object **body = ARRAY_DATA(ob, Object*);
             int len = ARRAY_LEN(ob);
             int i;
             TRACE_COMPACT("Scanning Array object @%p class is %s len is %d\n", ob, cb->name, len);
@@ -1093,19 +1098,18 @@ int threadChildren(Object *ob, Object *new_addr) {
             TRACE_COMPACT("Array object @%p class is %s - not Scanning...\n", ob, cb->name);
         }
     } else {
-        Object **body = (Object**)INST_DATA(ob);
         int i;
 
         if(IS_CLASS_CLASS(cb)) {
             TRACE_COMPACT("Found class object @%p name is %s\n", ob, CLASS_CB(ob)->name);
-            threadClassData((Class*)ob, (Class*)new_addr);
+            threadClassData(ob, new_addr);
         } else
             if(IS_CLASS_LOADER(cb)) {
                 TRACE_COMPACT("Found class loader object @%p class %s\n", ob, cb->name);
                 threadLoaderClasses(ob);
             } else
                 if(IS_REFERENCE(cb)) {
-                    Object **referent = &body[ref_referent_offset];
+                    Object **referent = &INST_DATA(ob, Object*, ref_referent_offset);
 
                     if(*referent != NULL) {
                         int ref_mark = IS_MARKED(*referent);
@@ -1121,14 +1125,14 @@ int threadChildren(Object *ob, Object *new_addr) {
                                 goto out;
 
                             TRACE_GC("Clearing the referent field.\n");
-                            *referent = 0;
+                            *referent = NULL;
                             cleared = TRUE;
                         }
 
                         /* If the reference has a queue, add it to the list for enqueuing
                            by the Reference Handler thread. */
 
-                        if(body[ref_queue_offset] != NULL) {
+                        if(INST_DATA(ob, Object*, ref_queue_offset) != NULL) {
                             TRACE_GC("Adding to list for enqueuing.\n");
 
                             ADD_TO_OBJECT_LIST(reference, new_addr);
@@ -1150,18 +1154,18 @@ out:
             int offset = cb->refs_offsets_table[i].start;
             int end = cb->refs_offsets_table[i].end;
 
-            for(; offset < end; offset++) {
-                Object **ob = (Object **)&body[offset];
-                TRACE_COMPACT("Offset %d reference @%p\n", offset, *ob);
+            for(; offset < end; offset += sizeof(Object*)) {
+                Object **ref = &INST_DATA(ob, Object*, offset);
+                TRACE_COMPACT("Offset %d reference @%p\n", offset, *ref);
 
-                if(*ob != NULL)
-                    threadReference(ob);
+                if(*ref != NULL)
+                    threadReference(ref);
             }
         }
     }
 
     /* Finally thread the object's class reference */
-    threadReference((Object**)&ob->class);
+    threadReference(&ob->class);
 
     return cleared;
 }
@@ -1447,7 +1451,7 @@ unsigned long gc0(int mark_soft_refs, int compact) {
         scan_time = endTime(&start)/1000000.0;
 
         jam_printf("<GC: Mark took %f seconds, %s took %f seconds>\n",
-                                  mark_time, compact ? "compact" : "scan", scan_time);
+                           mark_time, compact ? "compact" : "scan", scan_time);
     } else {
         doMark(self, mark_soft_refs);
         largest = compact ? doCompact() : doSweep(self);
@@ -1644,7 +1648,7 @@ void initialiseGC(InitArgs *args) {
     createVMThread("Reference Handler", referenceHandlerThreadLoop);
 
     /* Create and start VM thread for asynchronous GC */
-    if(!args->noasyncgc)
+    if(args->asyncgc)
         createVMThread("Async GC", asyncGCThreadLoop);
 
     /* GC will use mark-sweep or mark-compact as appropriate, but this
@@ -1858,8 +1862,7 @@ got_it:
 
 Object *allocObject(Class *class) {
     ClassBlock *cb = CLASS_CB(class);
-    int size = cb->object_size * sizeof(uintptr_t);
-    Object *ob = (Object *)gcMalloc(size + sizeof(Object));
+    Object *ob = gcMalloc(cb->object_size);
 
     if(ob != NULL) {
         ob->class = class;
@@ -1886,113 +1889,46 @@ Object *allocArray(Class *class, int size, int el_size) {
     Object *ob;
 
     /* Special check to protect against integer overflow */
-    if(size > (INT_MAX - sizeof(u4) - sizeof(Object)) / el_size) {
+    if(size > (INT_MAX - sizeof(uintptr_t) - sizeof(Object)) / el_size) {
         signalException(java_lang_OutOfMemoryError, NULL);
         return NULL;
     }
 
-    ob = (Object *)gcMalloc(size * el_size + sizeof(u4) + sizeof(Object));
+    ob = gcMalloc(size * el_size + sizeof(uintptr_t) + sizeof(Object));
 
     if(ob != NULL) {
         ob->class = class;
         ARRAY_LEN(ob) = size;
-        TRACE_ALLOC("<ALLOC: allocated %s array object @%p>\n", CLASS_CB(class)->name, ob);
+        TRACE_ALLOC("<ALLOC: allocated %s array object @%p>\n",
+                    CLASS_CB(class)->name, ob);
     }
 
     return ob;
 }
 
 Object *allocTypeArray(int type, int size) {
-    Class *class;
-    int el_size;
+    static char *array_names[] = {"[Z", "[C", "[F", "[D", "[B",
+                                  "[S", "[I", "[J"};
+    static int element_sizes[] = {1, 2, 4, 8, 1, 2, 4, 8};
+    static Class *array_classes[8];
 
+    int idx = type - T_BOOLEAN;
+     
     if(size < 0) {
         signalException(java_lang_NegativeArraySizeException, NULL);
         return NULL;
     }
 
-    switch(type) {
-        case T_BOOLEAN:
-            if(bool_array_class == NULL) {
-                bool_array_class = findArrayClass("[Z");
-                registerStaticClassRefLocked(&bool_array_class);
-            }
-            class = bool_array_class;
-            el_size = 1;
-            break;
+    if(array_classes[idx] == NULL) {
+        Class *class = findArrayClass(array_names[idx]);
 
-        case T_BYTE:
-            if(byte_array_class == NULL) {
-                byte_array_class = findArrayClass("[B");
-                registerStaticClassRefLocked(&byte_array_class);
-            }
-            class = byte_array_class;
-            el_size = 1;
-            break;
+        if(class == NULL)
+            return NULL;
 
-        case T_CHAR:
-            if(char_array_class == NULL) {
-                char_array_class = findArrayClass("[C");
-                registerStaticClassRefLocked(&char_array_class);
-            }
-            class = char_array_class;
-            el_size = 2;
-            break;
-
-        case T_SHORT:
-            if(short_array_class == NULL) {
-                short_array_class = findArrayClass("[S");
-                registerStaticClassRefLocked(&short_array_class);
-            }
-            class = short_array_class;
-            el_size = 2;
-            break;
-
-        case T_INT:
-            if(int_array_class == NULL) {
-                int_array_class = findArrayClass("[I");
-                registerStaticClassRefLocked(&int_array_class);
-            }
-            class = int_array_class;
-            el_size = 4;
-            break;
-
-        case T_FLOAT:
-            if(float_array_class == NULL) {
-                float_array_class = findArrayClass("[F");
-                registerStaticClassRefLocked(&float_array_class);
-            }
-            class = float_array_class;
-            el_size = 4;
-            break;
-
-        case T_DOUBLE:
-            if(double_array_class == NULL) {
-                double_array_class = findArrayClass("[D");
-                registerStaticClassRefLocked(&double_array_class);
-            }
-            class = double_array_class;
-            el_size = 8;
-            break;
-
-        case T_LONG:
-            if(long_array_class == NULL) {
-                long_array_class = findArrayClass("[J");
-                registerStaticClassRefLocked(&long_array_class);
-            }
-            class = long_array_class;
-            el_size = 8;
-            break;
-
-        default:
-            jam_printf("Invalid array type %d - aborting VM...\n", type);
-            exit(0);
+        registerStaticClassRefLocked(&array_classes[idx], class);
     }
 
-    if(class == NULL)
-        return NULL;
-
-    return allocArray(class, size, el_size);
+    return allocArray(array_classes[idx], size, element_sizes[idx]);
 }
 
 Object *allocMultiArray(Class *array_class, int dim, intptr_t *count) {
@@ -2010,7 +1946,8 @@ Object *allocMultiArray(Class *array_class, int dim, intptr_t *count) {
         if(array == NULL)
             return NULL;
 
-        body = ARRAY_DATA(array);
+        body = ARRAY_DATA(array, Object*);
+
         for(i = 0; i < *count; i++)
             if((*body++ = allocMultiArray(aclass, dim - 1, count + 1)) == NULL)
                 return NULL;
@@ -2048,7 +1985,7 @@ Object *allocMultiArray(Class *array_class, int dim, intptr_t *count) {
 }
 
 Class *allocClass() {
-    Class *class = (Class*)gcMalloc(sizeof(ClassBlock)+sizeof(Class));
+    Class *class = gcMalloc(sizeof(ClassBlock)+sizeof(Class));
 
     if(class != NULL) {
         SET_SPECIAL_OB(class);
@@ -2059,16 +1996,16 @@ Class *allocClass() {
 }
 
 Object *cloneObject(Object *ob) {
+    Object *clone;
     uintptr_t hdr = *HDR_ADDRESS(ob);
     int size = HDR_SIZE(hdr)-HEADER_SIZE;
-    Object *clone;
 
     /* If the object stores its original address the actual object
        data size will be smaller than the header size */
     if(HDR_HAS_HASHCODE(hdr))
         size -= OBJECT_GRAIN;
 
-    clone = (Object*)gcMalloc(size);
+    clone = gcMalloc(size);
 
     if(clone != NULL) {
         memcpy(clone, ob, size);
@@ -2085,7 +2022,7 @@ Object *cloneObject(Object *ob) {
             /* Safety.  If it's a classloader, clear native
                pointer to class table */
             if(IS_CLASS_LOADER(CLASS_CB(clone->class)))
-                INST_DATA(clone)[ldr_vmdata_offset] = 0;
+                INST_DATA(clone, Object*, ldr_vmdata_offset) = NULL;
         }
 
         TRACE_ALLOC("<ALLOC: cloned object @%p clone @%p>\n", ob, clone);
