@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2003, 2004, 2005, 2006, 2007, 2008
+ * Copyright (C) 2003, 2004, 2005, 2006, 2007, 2008, 2009
  * Robert Lougher <rob@lougher.org.uk>.
  *
  * This file is part of JamVM.
@@ -40,7 +40,7 @@
 
 #define PREPARE(ptr) ptr
 #define SCAVENGE(ptr) FALSE
-#define FOUND(ptr) ptr
+#define FOUND(ptr1, ptr2) ptr2
 
 static int verbose;
 static char *bootpath;
@@ -63,9 +63,22 @@ int ref_queue_offset;
 /* Cached offset of vmdata field in java.lang.ClassLoader objects */
 int ldr_vmdata_offset = -1;
 
+/* Helper method to create a Package Object representing a
+   package loaded by the boot loader */
+static MethodBlock *vm_loader_create_package = NULL;
+static Class *package_array_class;
 
-static Class *loader_data_class = NULL;
-static MethodBlock *ldr_new_unloader;
+/* hash table containing packages loaded by the boot loader */
+#define PCKG_INITSZE 1<<6
+static HashTable boot_packages;
+
+/* Hashtable entry for each package defined by the boot loader */
+typedef struct package_entry {
+    int index;
+    char name[0];
+} PackageEntry;
+
+static MethodBlock *ldr_new_unloader = NULL;
 static int ldr_data_tbl_offset;
 
 /* Instance of java.lang.Class for java.lang.Class */
@@ -83,12 +96,11 @@ int enqueue_mtbl_idx;
 
 /* hash table containing classes loaded by the boot loader and
    internally created arrays */
-
-#define INITSZE 1<<8
-static HashTable loaded_classes;
+#define CLASS_INITSZE 1<<8
+static HashTable boot_classes;
 
 /* Array large enough to hold all primitive classes -
- * access protected by loaded_classes hash table lock */
+ * access protected by boot_classes hash table lock */
 #define MAX_PRIM_CLASSES 9
 static Class *prim_classes[MAX_PRIM_CLASSES];
 
@@ -105,31 +117,31 @@ static Class *addClassToHash(Class *class, Object *class_loader) {
             CLASS_CB((Class *)ptr1)->name == CLASS_CB((Class *)ptr2)->name
 
     if(class_loader == NULL)
-        table = &loaded_classes;
+        table = &boot_classes;
     else {
-        Object *vmdata = (Object*)INST_DATA(class_loader)[ldr_vmdata_offset];
+        Object *vmdata = INST_DATA(class_loader, Object*, ldr_vmdata_offset);
 
         if(vmdata == NULL) {
             objectLock(class_loader);
-            vmdata = (Object*)INST_DATA(class_loader)[ldr_vmdata_offset];
+            vmdata = INST_DATA(class_loader, Object*, ldr_vmdata_offset);
 
             if(vmdata == NULL) {
-                if((vmdata = allocObject(loader_data_class)) == NULL) {
+                if((vmdata = allocObject(ldr_new_unloader->class)) == NULL) {
                     objectUnlock(class_loader);
                     return NULL;
                 }
 
                 table = sysMalloc(sizeof(HashTable));
-                initHashTable((*table), INITSZE, TRUE);
+                initHashTable((*table), CLASS_INITSZE, TRUE);
 
-                INST_DATA(vmdata)[ldr_data_tbl_offset] = (uintptr_t)table;
-                INST_DATA(class_loader)[ldr_vmdata_offset] = (uintptr_t)vmdata;
+                INST_DATA(vmdata, HashTable*, ldr_data_tbl_offset) = table;
+                INST_DATA(class_loader, Object*, ldr_vmdata_offset) = vmdata;
 
                 objectUnlock(class_loader);
             }
         }
 
-        table = (HashTable*)INST_DATA(vmdata)[ldr_data_tbl_offset];
+        table = INST_DATA(vmdata, HashTable*, ldr_data_tbl_offset);
     }
 
     /* Add if absent, no scavenge, locked */
@@ -151,10 +163,12 @@ static void prepareClass(Class *class) {
     }
 }
 
-Class *defineClass(char *classname, char *data, int offset, int len, Object *class_loader) {
-    unsigned char *ptr = (unsigned char *)data+offset;
-    int cp_count, intf_count, i;
+Class *defineClass(char *classname, char *data, int offset, int len,
+                   Object *class_loader) {
+
     u2 major_version, minor_version, this_idx, super_idx;
+    unsigned char *ptr = (unsigned char *)data + offset;
+    int cp_count, intf_count, i;
     u2 attr_count;
     u4 magic;
 
@@ -253,7 +267,8 @@ Class *defineClass(char *classname, char *data, int offset, int len, Object *cla
            }
 
            default:
-               signalException(java_lang_ClassFormatError, "bad constant pool tag");
+               signalException(java_lang_ClassFormatError,
+                               "bad constant pool tag");
                return NULL;
         }
     }
@@ -268,7 +283,8 @@ Class *defineClass(char *classname, char *data, int offset, int len, Object *cla
     classblock->name = CP_UTF8(constant_pool, CP_CLASS(constant_pool, this_idx));
 
     if(classname && strcmp(classblock->name, classname) != 0) {
-        signalException(java_lang_NoClassDefFoundError, "class file has wrong name");
+        signalException(java_lang_NoClassDefFoundError,
+                        "class file has wrong name");
         return NULL;
     }
 
@@ -582,7 +598,7 @@ Class *createArrayClass(char *classname, Object *class_loader) {
     classblock->method_table = CLASS_CB(classblock->super)->method_table;
 
     classblock->interfaces_count = 2;
-    classblock->interfaces = (Class**)sysMalloc(2*sizeof(Class*));
+    classblock->interfaces = sysMalloc(sizeof(Class*) * 2);
     classblock->interfaces[0] = findSystemClass0(SYMBOL(java_lang_Cloneable));
     classblock->interfaces[1] = findSystemClass0(SYMBOL(java_io_Serializable));
 
@@ -592,7 +608,8 @@ Class *createArrayClass(char *classname, Object *class_loader) {
        this is used to speed up type checking (instanceof) */
 
     if(classname[1] == '[') {
-        Class *comp_class = findArrayClassFromClassLoader(classname + 1, class_loader);
+        Class *comp_class = findArrayClassFromClassLoader(classname + 1,
+                                                          class_loader);
 
         if(comp_class == NULL)
             goto error;
@@ -601,11 +618,13 @@ Class *createArrayClass(char *classname, Object *class_loader) {
         classblock->dim = CLASS_CB(comp_class)->dim + 1;
     } else { 
         if(classname[1] == 'L') {
-            char element_name[len-2];
+            char element_name[len - 2];
 
-            strcpy(element_name, classname + 2);
-            element_name[len-3] = '\0';
-            classblock->element_class = findClassFromClassLoader(element_name, class_loader);
+            memcpy(element_name, classname + 2, len - 3);
+            element_name[len - 3] = '\0';
+
+            classblock->element_class = findClassFromClassLoader(element_name,
+                                                                 class_loader);
         } else
             classblock->element_class = findPrimitiveClass(classname[1]);
 
@@ -637,8 +656,7 @@ error:
     return found;
 }
 
-Class *
-createPrimClass(char *classname, int index) {
+Class *createPrimClass(char *classname, int index) {
     Class *class;
     ClassBlock *classblock;
  
@@ -652,10 +670,10 @@ createPrimClass(char *classname, int index) {
 
     prepareClass(class);
 
-    lockHashTable(loaded_classes);
+    lockHashTable(boot_classes);
     if(prim_classes[index] == NULL)
         prim_classes[index] = class;
-    unlockHashTable(loaded_classes);
+    unlockHashTable(boot_classes);
 
     if(verbose)
         jam_printf("[Created primitive class %s]\n", classname);
@@ -663,6 +681,158 @@ createPrimClass(char *classname, int index) {
     return prim_classes[index];
 }
 
+/* Layout the instance data.
+
+   The object layout places 64-bit fields on a double-word boundary
+   as on some architectures this leads to better performance (and
+   misaligned loads/store may cause traps).
+
+   Reference fields are also placed together as these must be scanned
+   during GC, and placing them together reduces the number of entries
+   required in the reference offsets list.
+
+   Double/long fields are layed out first, then references and finally
+   int-sized fields.  When padding is needed for 64-bit fields we try
+   to place an int-sized field, and only leave a hole when no int-sized
+   fields are available */
+
+void prepareFields(Class *class) {
+    ClassBlock *cb = CLASS_CB(class);
+    Class *super = (cb->access_flags & ACC_INTERFACE) ? NULL
+                                                      : cb->super;
+
+    RefsOffsetsEntry *spr_rfs_offsts_tbl = NULL;
+    int spr_rfs_offsts_sze = 0;
+
+    FieldBlock *ref_head = NULL;
+    FieldBlock *int_head = NULL;
+    FieldBlock *dbl_head = NULL;
+
+    int field_offset = sizeof(Object);
+    int refs_start_offset = 0;
+    int refs_end_offset = 0;
+    int i;
+
+    if(super != NULL) {
+        field_offset = CLASS_CB(super)->object_size;
+        spr_rfs_offsts_sze = CLASS_CB(super)->refs_offsets_size;
+        spr_rfs_offsts_tbl = CLASS_CB(super)->refs_offsets_table;
+    }
+
+    /* Initialise static fields to default value, and separate
+       instance fields into three linked lists, holding
+       int-sized fields, double-sized fields and reference
+       fields */
+
+    for(i = 0; i < cb->fields_count; i++) {
+        FieldBlock *fb = &cb->fields[i];
+
+        if(fb->access_flags & ACC_STATIC)
+            fb->u.static_value.l = 0;
+        else {
+            FieldBlock **list;
+
+            if(fb->type[0] == 'L' || fb->type[0] == '[')
+                list = &ref_head;
+            else
+                if(fb->type[0] == 'J' || fb->type[0] == 'D')
+                    list = &dbl_head;
+                else
+                    list = &int_head;
+
+            fb->u.static_value.p = *list;
+            *list = fb;
+        }
+
+        fb->class = class;
+    }
+
+    /* Layout the double-sized fields.  If padding is required,
+       use the first int-sized field (int_list head), or leave
+       a hole if no int-fields */
+
+    if(dbl_head != NULL) {
+        if(field_offset & 0x7) {
+            if(int_head != NULL) {
+                FieldBlock *fb = int_head;
+                int_head = int_head->u.static_value.p;
+                fb->u.offset = field_offset;
+            }
+            field_offset += 4;
+        }
+
+        do {
+            FieldBlock *fb = dbl_head;
+            dbl_head = dbl_head->u.static_value.p;
+            fb->u.offset = field_offset;
+            field_offset += 8;
+        } while(dbl_head != NULL);
+    }
+
+    /* Layout the reference fields.  If padding is required,
+       use an int-sized field (int_list head), or leave
+       a hole if no int-fields remaining */
+
+    if(ref_head != NULL) {
+        if(sizeof(Object*) == 8 && field_offset & 0x7) {
+            if(int_head != NULL) {
+                FieldBlock *fb = int_head;
+                int_head = int_head->u.static_value.p;
+                fb->u.offset = field_offset;
+            }
+            field_offset += 4;
+        }
+
+        refs_start_offset = field_offset;
+
+        do {
+            FieldBlock *fb = ref_head;
+            ref_head = ref_head->u.static_value.p;
+            fb->u.offset = field_offset;
+            field_offset += sizeof(Object*);
+        } while(ref_head != NULL);
+
+        refs_end_offset = field_offset;
+    }
+
+    /* Layout the remaining int-sized fields */
+
+    while(int_head != NULL) {
+        FieldBlock *fb = int_head;
+        int_head = int_head->u.static_value.p;
+        fb->u.offset = field_offset;
+        field_offset += 4;
+    }
+
+   cb->object_size = field_offset;
+
+   /* Construct the reference offsets list.  This is used to speed up
+      scanning of an objects references during the mark phase of GC.
+      If possible, merge the entry with the previous entry */
+
+   if(refs_start_offset) {
+       if(spr_rfs_offsts_sze > 0 && spr_rfs_offsts_tbl[spr_rfs_offsts_sze-1].end
+                                           == refs_start_offset) {
+
+           cb->refs_offsets_size = spr_rfs_offsts_sze;
+           refs_start_offset = spr_rfs_offsts_tbl[spr_rfs_offsts_sze-1].start;
+       } else
+           cb->refs_offsets_size = spr_rfs_offsts_sze + 1;
+
+      cb->refs_offsets_table = sysMalloc(cb->refs_offsets_size *
+                                         sizeof(RefsOffsetsEntry));
+
+      memcpy(cb->refs_offsets_table, spr_rfs_offsts_tbl,
+             spr_rfs_offsts_sze * sizeof(RefsOffsetsEntry));
+
+      cb->refs_offsets_table[cb->refs_offsets_size-1].start = refs_start_offset;
+      cb->refs_offsets_table[cb->refs_offsets_size-1].end = refs_end_offset;
+   } else {
+       cb->refs_offsets_size = spr_rfs_offsts_sze;
+       cb->refs_offsets_table = spr_rfs_offsts_tbl;
+   }
+}
+
 #define MRNDA_CACHE_SZE 10
 
 #define resizeMTable(method_table, method_table_size, miranda, count)  \
@@ -692,7 +862,6 @@ void linkClass(Class *class) {
    ClassBlock *cb = CLASS_CB(class);
    Class *super = (cb->access_flags & ACC_INTERFACE) ? NULL : cb->super;
 
-   RefsOffsetsEntry *spr_rfs_offsts_tbl = NULL;
    ITableEntry *spr_imthd_tbl = NULL;
    MethodBlock **method_table = NULL;
    MethodBlock **spr_mthd_tbl = NULL;
@@ -700,18 +869,14 @@ void linkClass(Class *class) {
    MethodBlock *mb;
    FieldBlock *fb;
 
-   int spr_rfs_offsts_sze = 0;
    int new_methods_count = 0;
    int spr_imthd_tbl_sze = 0;
    int itbl_offset_count = 0;
    int spr_mthd_tbl_sze = 0;
    int method_table_size;
    int new_itable_count;
-   int spr_obj_sze = 0;
-   int refs_end_offset;
    int itbl_idx, i, j;
    int spr_flags = 0;
-   int field_offset;
 
    if(cb->state >= CLASS_LINKED)
        return;
@@ -730,48 +895,15 @@ void linkClass(Class *class) {
           linkClass(super);
 
       spr_flags = super_cb->flags;
-      spr_obj_sze = super_cb->object_size;
       spr_mthd_tbl = super_cb->method_table;
       spr_imthd_tbl = super_cb->imethod_table;
       spr_mthd_tbl_sze = super_cb->method_table_size;
       spr_imthd_tbl_sze = super_cb->imethod_table_size;
-      spr_rfs_offsts_sze = super_cb->refs_offsets_size;
-      spr_rfs_offsts_tbl = super_cb->refs_offsets_table;
    }
 
-   /* prepare fields */
+   /* Calculate object layout */
 
-   field_offset = spr_obj_sze;
-
-   for(fb = cb->fields, i = 0; i < cb->fields_count; i++,fb++) {
-      if(fb->access_flags & ACC_STATIC) {
-         /* init to default value */
-         if((*fb->type == 'J') || (*fb->type == 'D'))
-            *(long long *)&fb->static_value = 0;
-         else
-            fb->static_value = 0;
-      } else {
-         /* calc field offset */
-         if((*fb->type == 'L') || (*fb->type == '['))
-            fb->offset = field_offset++;
-      }
-      fb->class = class;
-   }
-
-   refs_end_offset = field_offset;
-
-   for(fb = cb->fields, i = 0; i < cb->fields_count; i++,fb++)
-      if(!(fb->access_flags & ACC_STATIC) &&
-                 (*fb->type != 'L') && (*fb->type != '[')) {
-         /* calc field offset */
-         fb->offset = field_offset;
-         if((*fb->type == 'J') || (*fb->type == 'D'))
-             field_offset += 2;
-         else
-             field_offset += 1;
-      }
-
-   cb->object_size = field_offset;
+   prepareFields(class);
 
    /* prepare methods */
 
@@ -1024,13 +1156,14 @@ void linkClass(Class *class) {
        }
 
        for(fb = cb->fields, i = 0; i < cb->fields_count; i++,fb++)
-           if(fb->offset > ref_fb->offset)
-               fb->offset--;
+           if(fb->u.offset > ref_fb->u.offset)
+               fb->u.offset -= sizeof(Object*);
+
+       ref_referent_offset = ref_fb->u.offset = cb->object_size - sizeof(Object*);
+       cb->refs_offsets_table[cb->refs_offsets_size-1].end -= sizeof(Object*);
 
-       ref_referent_offset = ref_fb->offset = field_offset - 1;
        enqueue_mtbl_idx = enqueue_mb->method_table_index;
-       ref_queue_offset = queue_fb->offset;
-       refs_end_offset--;
+       ref_queue_offset = queue_fb->u.offset;
 
        cb->flags |= REFERENCE;
    }
@@ -1056,36 +1189,10 @@ void linkClass(Class *class) {
            exitVM(1);
        }
 
-       ldr_vmdata_offset = ldr_fb->offset;
+       ldr_vmdata_offset = ldr_fb->u.offset;
        cb->flags |= CLASS_LOADER;
    }
 
-   /* Construct the reference offsets list.  This is used to speed up
-      scanning of an objects references during the mark phase of GC. */
-
-   if(refs_end_offset > spr_obj_sze) {
-       int new_start;
-
-       if(spr_rfs_offsts_sze > 0 && spr_rfs_offsts_tbl[spr_rfs_offsts_sze-1].end == spr_obj_sze) {
-           cb->refs_offsets_size = spr_rfs_offsts_sze;
-           new_start = spr_rfs_offsts_tbl[spr_rfs_offsts_sze-1].start;
-       } else {
-           cb->refs_offsets_size = spr_rfs_offsts_sze + 1;
-           new_start = spr_obj_sze;
-      }
-
-      cb->refs_offsets_table = sysMalloc(cb->refs_offsets_size * sizeof(RefsOffsetsEntry));
-
-      memcpy(cb->refs_offsets_table, spr_rfs_offsts_tbl,
-                                     spr_rfs_offsts_sze * sizeof(RefsOffsetsEntry));
-
-      cb->refs_offsets_table[cb->refs_offsets_size-1].start = new_start;
-      cb->refs_offsets_table[cb->refs_offsets_size-1].end = refs_end_offset;
-   } else {
-       cb->refs_offsets_size = spr_rfs_offsts_sze;
-       cb->refs_offsets_table = spr_rfs_offsts_tbl;
-   }
-
    cb->state = CLASS_LINKED;
 
 unlock:
@@ -1151,9 +1258,9 @@ Class *initClass(Class *class) {
    for(i = 0; i < cb->fields_count; i++,fb++)
       if((fb->access_flags & ACC_STATIC) && fb->constant) {
          if((*fb->type == 'J') || (*fb->type == 'D'))
-            *(u8*)&fb->static_value = *(u8*)&(CP_INFO(cp, fb->constant));
+            fb->u.static_value.l = *(u8*)&(CP_INFO(cp, fb->constant));
          else
-            fb->static_value = resolveSingleConstant(class, fb->constant);
+            fb->u.static_value.u = resolveSingleConstant(class, fb->constant);
       }
 
    if((mb = findMethod(class, SYMBOL(class_init), SYMBOL(___V))) != NULL)
@@ -1216,6 +1323,31 @@ char *findFileEntry(char *path, int *file_len) {
     return NULL;
 }
 
+void defineBootPackage(char *classname, int index) {
+    char *last_slash = strrchr(classname, '/');
+
+    if(last_slash != NULL && last_slash != classname) {
+        int len = last_slash - classname + 1;
+        PackageEntry *package = sysMalloc(sizeof(PackageEntry) + len);
+        PackageEntry *hashed;
+        
+        package->index = index;
+        slash2dots2buff(classname, package->name, len);
+
+#undef HASH
+#undef COMPARE
+#define HASH(ptr) utf8Hash(((PackageEntry*)ptr)->name)
+#define COMPARE(ptr1, ptr2, hash1, hash2) (hash1 == hash2 && \
+            utf8Comp(((PackageEntry*)ptr1)->name, ((PackageEntry*)ptr2)->name))
+
+        /* Add if absent, no scavenge, locked */
+        findHashEntry(boot_packages, package, hashed, TRUE, FALSE, TRUE);
+
+        if(package != hashed)
+            sysFree(package);
+    }
+}
+
 Class *loadSystemClass(char *classname) {
     int file_len, fname_len = strlen(classname) + 8;
     char buff[max_cp_element_len + fname_len];
@@ -1229,15 +1361,19 @@ Class *loadSystemClass(char *classname) {
 
     for(i = 0; i < bcp_entries && data == NULL; i++)
         if(bootclasspath[i].zip)
-            data = findArchiveEntry(filename+1, bootclasspath[i].zip, &file_len);
+            data = findArchiveEntry(filename + 1, bootclasspath[i].zip,
+                                    &file_len);
         else
-            data = findFileEntry(strcat(strcpy(buff, bootclasspath[i].path), filename), &file_len);
+            data = findFileEntry(strcat(strcpy(buff, bootclasspath[i].path),
+                                 filename), &file_len);
 
     if(data == NULL) {
         signalException(java_lang_NoClassDefFoundError, classname);
         return NULL;
     }
 
+    defineBootPackage(classname, i - 1);
+
     class = defineClass(classname, data, 0, file_len, NULL);
     sysFree(data);
 
@@ -1267,14 +1403,14 @@ Class *findHashedClass(char *classname, Object *class_loader) {
         return NULL;
 
     if(class_loader == NULL)
-        table = &loaded_classes;
+        table = &boot_classes;
     else {
-        Object *vmdata = (Object*)INST_DATA(class_loader)[ldr_vmdata_offset];
+        Object *vmdata = INST_DATA(class_loader, Object*, ldr_vmdata_offset);
 
         if(vmdata == NULL)
             return NULL;
 
-        table = (HashTable*)INST_DATA(vmdata)[ldr_data_tbl_offset];
+        table = INST_DATA(vmdata, HashTable*, ldr_data_tbl_offset);
     }
 
 #undef HASH
@@ -1327,31 +1463,40 @@ Class *findPrimitiveClass(char prim_type) {
 
    switch(prim_type) {
       case 'Z':
-          classname = SYMBOL(boolean); index = 1;
+          classname = SYMBOL(boolean);
+          index = PRIM_IDX_BOOLEAN;
           break;
       case 'B':
-          classname = SYMBOL(byte); index = 2;
+          classname = SYMBOL(byte);
+          index = PRIM_IDX_BYTE;
           break;
       case 'C':
-          classname = SYMBOL(char); index = 3;
+          classname = SYMBOL(char);
+          index = PRIM_IDX_CHAR;
           break;
       case 'S':
-          classname = SYMBOL(short); index = 4;
+          classname = SYMBOL(short);
+           index = PRIM_IDX_SHORT;
           break;
       case 'I':
-          classname = SYMBOL(int); index = 5;
+          classname = SYMBOL(int);
+           index = PRIM_IDX_INT;
           break;
       case 'F':
-          classname = SYMBOL(float); index = 6;
+          classname = SYMBOL(float);
+           index = PRIM_IDX_FLOAT;
           break;
       case 'J':
-          classname = SYMBOL(long); index = 7;
+          classname = SYMBOL(long);
+           index = PRIM_IDX_LONG;
           break;
       case 'D':
-          classname = SYMBOL(double); index = 8;
+          classname = SYMBOL(double);
+           index = PRIM_IDX_DOUBLE;
           break;
       case 'V':
-          classname = SYMBOL(void); index = 0;
+          classname = SYMBOL(void);
+           index = PRIM_IDX_VOID;
           break;
       default:
           signalException(java_lang_NoClassDefFoundError, NULL);
@@ -1432,14 +1577,73 @@ Object *getSystemClassLoader() {
     return NULL;
 }
 
+Object *createBootPackage(PackageEntry *package_entry) {
+    Object *name = createString(package_entry->name);
+
+    if(name != NULL) {
+        Object *package = *(Object**)executeStaticMethod(
+                                            vm_loader_create_package->class,
+                                            vm_loader_create_package, name,
+                                            package_entry->index);
+
+        if(!exceptionOccurred()) 
+            return package;
+    }
+
+    return NULL;
+}
+
+Object *bootPackage(char *package_name) {
+    PackageEntry *hashed;
+
+#undef HASH
+#undef COMPARE
+#define HASH(ptr) utf8Hash(ptr)
+#define COMPARE(ptr1, ptr2, hash1, hash2) (hash1 == hash2 && \
+                                 utf8Comp(ptr1, ((PackageEntry*)ptr2)->name))
+
+    /* Do not add if absent, no scavenge, locked */
+    findHashEntry(boot_packages, package_name, hashed, FALSE, FALSE, TRUE);
+
+    if(hashed != NULL)
+        return createBootPackage(hashed);
+
+    return NULL;
+}
+
+#define ITERATE(ptr)                                       \
+    if((data[--count] = createBootPackage(ptr)) == NULL) { \
+        array = NULL;                                      \
+        goto error;                                        \
+    }
+
+Object *bootPackages() {
+    Object **data, *array;
+    int count;
+
+    lockHashTable(boot_packages);
+
+    count = hashTableCount(boot_packages);
+    if((array = allocArray(package_array_class, count, sizeof(Object*))) == NULL)
+        goto error;
+
+    data = ARRAY_DATA(array, Object*);
+    hashIterate(boot_packages);
+
+error:
+    unlockHashTable(boot_packages);
+    return array;
+}
+
 /* gc support for marking classes */
 
-#define ITERATE(ptr)  markRoot(ptr)
+#undef ITERATE
+#define ITERATE(ptr) markRoot(ptr)
 
 void markBootClasses() {
    int i;
 
-   hashIterate(loaded_classes);
+   hashIterate(boot_classes);
 
    for(i = 0; i < MAX_PRIM_CLASSES; i++)
        if(prim_classes[i] != NULL)
@@ -1452,7 +1656,7 @@ void markBootClasses() {
 void threadBootClasses() {
    int i;
 
-   hashIterateP(loaded_classes);
+   hashIterateP(boot_classes);
 
    for(i = 0; i < MAX_PRIM_CLASSES; i++)
        if(prim_classes[i] != NULL)
@@ -1465,10 +1669,10 @@ void threadBootClasses() {
         markObject(ptr, mark, mark_soft_refs)
 
 void markLoaderClasses(Object *class_loader, int mark, int mark_soft_refs) {
-    Object *vmdata = (Object*)INST_DATA(class_loader)[ldr_vmdata_offset];
+    Object *vmdata = INST_DATA(class_loader, Object*, ldr_vmdata_offset);
 
     if(vmdata != NULL) {
-        HashTable *table = (HashTable*)INST_DATA(vmdata)[ldr_data_tbl_offset];
+        HashTable *table = INST_DATA(vmdata, HashTable*, ldr_data_tbl_offset);
         hashIterate((*table));
     }
 }
@@ -1477,10 +1681,10 @@ void markLoaderClasses(Object *class_loader, int mark, int mark_soft_refs) {
 #define ITERATE(ptr)  threadReference((Object**)ptr)
 
 void threadLoaderClasses(Object *class_loader) {
-    Object *vmdata = (Object*)INST_DATA(class_loader)[ldr_vmdata_offset];
+    Object *vmdata = INST_DATA(class_loader, Object*, ldr_vmdata_offset);
 
     if(vmdata != NULL) {
-        HashTable *table = (HashTable*)INST_DATA(vmdata)[ldr_data_tbl_offset];
+        HashTable *table = INST_DATA(vmdata, HashTable*, ldr_data_tbl_offset);
         hashIterateP((*table));
     }
 }
@@ -1577,10 +1781,10 @@ void freeClassData(Class *class) {
 }
 
 void freeClassLoaderData(Object *class_loader) {
-    Object *vmdata = (Object*)INST_DATA(class_loader)[ldr_vmdata_offset];
+    Object *vmdata = INST_DATA(class_loader, Object*, ldr_vmdata_offset);
 
     if(vmdata != NULL) {
-        HashTable *table = (HashTable*)INST_DATA(vmdata)[ldr_data_tbl_offset];
+        HashTable *table = INST_DATA(vmdata, HashTable*, ldr_data_tbl_offset);
         gcFreeHashTable((*table));
         gcPendingFree(table);
     }
@@ -1591,7 +1795,7 @@ void freeClassLoaderData(Object *class_loader) {
    function, which will be called from the unloader finalizer
    when the class loader is garbage collected */
 void newLibraryUnloader(Object *class_loader, void *entry) {
-    Object *vmdata = (Object*)INST_DATA(class_loader)[ldr_vmdata_offset];
+    Object *vmdata = INST_DATA(class_loader, Object*, ldr_vmdata_offset);
     
     if(vmdata != NULL)
         executeMethod(vmdata, ldr_new_unloader, (long long)(uintptr_t)entry);
@@ -1653,9 +1857,13 @@ char *getClassPath() {
     return classpath;
 }
 
+#ifdef __linux__
+int filter(const struct dirent *entry) {
+#else
 int filter(struct dirent *entry) {
+#endif
     int len = strlen(entry->d_name);
-    char *ext = &entry->d_name[len-4];
+    char *ext = (char*)&entry->d_name[len-4];
 
     return len >= 4 && (strcasecmp(ext, ".zip") == 0 ||
                         strcasecmp(ext, ".jar") == 0);
@@ -1760,35 +1968,53 @@ int bootClassPathSize() {
 }
 
 Object *bootClassPathResource(char *filename, int index) {
-    if(index < bcp_entries) {
-        /* Alloc enough space for Jar file URL -- jar:file://<path>!/<filename> */
-        char buff[strlen(filename) + strlen(bootclasspath[index].path) + 14];
+    Object *res = NULL;
 
-        if(bootclasspath[index].zip) {
+    if(index < bcp_entries) {
+        char *buff, *path = bootclasspath[index].path;
+        int path_len = strlen(path);
+
+        if(path[0] != '/') {
+            char *cwd = getCwd();
+            path_len += strlen(cwd) + 1;
+            path = strcat(strcat(strcpy(sysMalloc(path_len + 1), cwd), "/"), path);
+        }
+
+        /* Alloc enough space for Jar file URL -- jar:file://<path>!/<filename> */
+        buff = sysMalloc(strlen(filename) + path_len + 14);
+
+        if(bootclasspath[index].zip != NULL) {
             while(*filename == '/')
                 filename++;
 
-            if(!findArchiveDirEntry(filename, bootclasspath[index].zip))
-                return NULL;
+            if(findArchiveDirEntry(filename, bootclasspath[index].zip) == NULL)
+                goto out;
 
-            sprintf(buff, "jar:file://%s!/%s", bootclasspath[index].path, filename);
+            sprintf(buff, "jar:file://%s!/%s", path, filename);
         } else {
             struct stat info;
 
-            sprintf(buff, "file://%s/%s", bootclasspath[index].path, filename);
+            sprintf(buff, "file://%s/%s", path, filename);
             if(stat(&buff[7], &info) != 0 || S_ISDIR(info.st_mode))
-                return NULL;
+                goto out;
         }
 
-        return createString(buff);
+        res = createString(buff);
+
+out:
+        if(path != bootclasspath[index].path)
+            sysFree(path);
+        sysFree(buff);
     }
 
-    return NULL;
+    return res;
 }
 
 void initialiseClass(InitArgs *args) {
     char *bcp = setBootClassPath(args->bootpath, args->bootpathopt);
     FieldBlock *hashtable = NULL;
+    Class *loader_data_class;
+    Class *vm_loader_class;
 
     if(!(bcp && parseBootClassPath(bcp))) {
         jam_fprintf(stderr, "bootclasspath is empty!\n");
@@ -1799,22 +2025,41 @@ void initialiseClass(InitArgs *args) {
     setClassPath(args->classpath);
 
     /* Init hash table, and create lock */
-    initHashTable(loaded_classes, INITSZE, TRUE);
+    initHashTable(boot_classes, CLASS_INITSZE, TRUE);
+
+    initHashTable(boot_packages, PCKG_INITSZE, TRUE);
 
     loader_data_class = findSystemClass0(SYMBOL(jamvm_java_lang_VMClassLoaderData));
-    registerStaticClassRef(&loader_data_class);
-
     if(loader_data_class != NULL) {
         ldr_new_unloader = findMethod(loader_data_class, SYMBOL(newLibraryUnloader),
                                                          SYMBOL(_J__V));
-        hashtable = findField(loader_data_class, SYMBOL(hashtable), SYMBOL(I));
+        hashtable = findField(loader_data_class, SYMBOL(hashtable), SYMBOL(J));
     }
 
     if(hashtable == NULL || ldr_new_unloader == NULL) {
         jam_fprintf(stderr, "Fatal error: Bad VMClassLoaderData (missing or corrupt)\n");
         exitVM(1);
     }
-    ldr_data_tbl_offset = hashtable->offset;
+    ldr_data_tbl_offset = hashtable->u.offset;
+
+    vm_loader_class = findSystemClass0(SYMBOL(java_lang_VMClassLoader));
+    if(vm_loader_class != NULL)
+       vm_loader_create_package =
+                  findMethod(vm_loader_class, SYMBOL(createBootPackage),
+                             SYMBOL(_java_lang_String_I__java_lang_Package));
+
+    if(vm_loader_create_package == NULL) {
+        jam_fprintf(stderr, "Fatal error: Bad java.lang.VMClassLoader (missing or corrupt)\n");
+        exitVM(1);
+    }
+
+    package_array_class = findArrayClass(SYMBOL(array_java_lang_Package));
+    registerStaticClassRef(&package_array_class);
+
+    if(package_array_class == NULL) {
+        jam_fprintf(stderr, "Fatal error: missing java.lang.Package\n");
+        exitVM(1);
+    }
 
     /* Register the address of where the java.lang.Class ref _will_ be */
     registerStaticClassRef(&java_lang_Class);
