@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2003, 2004, 2005, 2006, 2007
+ * Copyright (C) 2003, 2004, 2005, 2006, 2007, 2008
  * Robert Lougher <rob@lougher.org.uk>.
  *
  * This file is part of JamVM.
@@ -19,6 +19,10 @@
  * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  */
 
+/* Ensure operand stack is double-word aligned.  This leads to
+   better double floating-point performance */
+#define ALIGN_OSTACK(pntr) (uintptr_t*)(((uintptr_t)(pntr) + 7) & ~7)
+
 #define CREATE_TOP_FRAME(ee, class, mb, sp, ret)                \
 {                                                               \
     Frame *last = ee->last_frame;                               \
@@ -28,7 +32,7 @@
                                                                 \
     ret = (void*) (sp = (uintptr_t*)(dummy+1));                 \
     new_frame = (Frame *)(sp + mb->max_locals);                 \
-    new_ostack = (uintptr_t *)(new_frame + 1);                  \
+    new_ostack = ALIGN_OSTACK(new_frame + 1);                   \
                                                                 \
     if((char*)(new_ostack + mb->max_stack) > ee->stack_end) {   \
         if(ee->overflow++) {                                    \
