@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2003, 2004, 2005, 2006, 2007, 2008
+ * Copyright (C) 2003, 2004, 2005, 2006, 2007, 2008, 2009
  * Robert Lougher <rob@lougher.org.uk>.
  *
  * This file is part of JamVM.
@@ -27,6 +27,7 @@
 #include "jam.h"
 #include "class.h"
 #include "symbol.h"
+#include "excep.h"
 
 #ifdef USE_ZIP
 #define BCP_MESSAGE "<jar/zip files and directories separated by :>"
@@ -45,7 +46,7 @@ void showNonStandardOptions() {
     printf("\t\t   locations where to find Classpath's classes\n");
     printf("  -Xbootclasspath/v:%s\n", BCP_MESSAGE);
     printf("\t\t   locations where to find JamVM's classes\n");
-    printf("  -Xnoasyncgc\t   turn off asynchronous garbage collection\n");
+    printf("  -Xasyncgc\t   turn on asynchronous garbage collection\n");
     printf("  -Xcompactalways  always compact the heap when garbage-collecting\n");
     printf("  -Xnocompact\t   turn off heap-compaction\n");
 #ifdef INLINING
@@ -57,9 +58,12 @@ void showNonStandardOptions() {
     printf("\t\t   <value> copy when usage reaches threshold value\n");
     printf("  -Xcodemem:[unlimited|<size>] (default maximum heapsize/4)\n");
 #endif
-    printf("  -Xms<size>\t   set the initial size of the heap (default = %dM)\n", DEFAULT_MIN_HEAP/MB);
-    printf("  -Xmx<size>\t   set the maximum size of the heap (default = %dM)\n", DEFAULT_MAX_HEAP/MB);
-    printf("  -Xss<size>\t   set the Java stack size for each thread (default = %dK)\n", DEFAULT_STACK/KB);
+    printf("  -Xms<size>\t   set the initial size of the heap "
+           "(default = %dM)\n", DEFAULT_MIN_HEAP/MB);
+    printf("  -Xmx<size>\t   set the maximum size of the heap "
+           "(default = %dM)\n", DEFAULT_MAX_HEAP/MB);
+    printf("  -Xss<size>\t   set the Java stack size for each thread "
+           "(default = %dK)\n", DEFAULT_STACK/KB);
     printf("\t\t   size may be followed by K,k or M,m (e.g. 2M)\n");
 }
 
@@ -69,24 +73,27 @@ void showUsage(char *name) {
     printf("   or  %s [-options] -jar jarfile [arg1 arg2 ...]\n", name);
     printf("                 (to run a standalone jar file)\n");
     printf("\nwhere options include:\n");
-    printf("  -help\t\t   print out this message\n");
-    printf("  -version\t   print out version number and copyright information\n");
-    printf("  -showversion     show version number and copyright and continue\n");
-    printf("  -fullversion     show jpackage-compatible version number and exit\n");
-    printf("  -cp -classpath   <jar/zip files and directories separated by :>\n");
+    printf("  -client\t   compatibility (ignored)\n");
+    printf("  -server\t   compatibility (ignored)\n\n");
+    printf("  -cp\t\t   <jar/zip files and directories separated by :>\n");
+    printf("  -classpath\t   <jar/zip files and directories separated by :>\n");
     printf("\t\t   locations where to find application classes\n");
+    printf("  -D<name>=<value> set a system property\n");
     printf("  -verbose[:class|gc|jni]\n");
     printf("\t\t   :class print out information about class loading, etc.\n");
     printf("\t\t   :gc print out results of garbage collection\n");
     printf("\t\t   :jni print out native method dynamic resolution\n");
-    printf("  -D<name>=<value> set a system property\n");
+    printf("  -version\t   print out version number and copyright information\n");
+    printf("  -showversion     show version number and copyright and continue\n");
+    printf("  -fullversion     show jpackage-compatible version number and exit\n");
+    printf("  -? -help\t   print out this message\n");
     printf("  -X\t\t   show help on non-standard options\n");
 }
 
 void showVersionAndCopyright() {
     printf("java version \"%s\"\n", JAVA_COMPAT_VERSION);
     printf("JamVM version %s\n", VERSION);
-    printf("Copyright (C) 2003-2008 Robert Lougher <rob@lougher.org.uk>\n\n");
+    printf("Copyright (C) 2003-2009 Robert Lougher <rob@lougher.org.uk>\n\n");
     printf("This program is free software; you can redistribute it and/or\n");
     printf("modify it under the terms of the GNU General Public License\n");
     printf("as published by the Free Software Foundation; either version 2,\n");
@@ -144,8 +151,10 @@ int parseCommandLine(int argc, char *argv[], InitArgs *args) {
             }
 
             if((args->props_count = props_count)) {
-                args->commandline_props = (Property*)sysMalloc(props_count * sizeof(Property));
-                memcpy(args->commandline_props, &props[0], props_count * sizeof(Property));
+                args->commandline_props = sysMalloc(props_count *
+                                                    sizeof(Property));
+                memcpy(args->commandline_props, &props[0], props_count *
+                                                           sizeof(Property));
             }
 
             if(is_jar) {
@@ -156,7 +165,8 @@ int parseCommandLine(int argc, char *argv[], InitArgs *args) {
             return i;
         }
 
-        if(strcmp(argv[i], "-help") == 0) {
+        if(strcmp(argv[i], "-?") == 0 ||
+           strcmp(argv[i], "-help") == 0) {
             status = 0;
             break;
 
@@ -190,41 +200,59 @@ int parseCommandLine(int argc, char *argv[], InitArgs *args) {
             else if(strcmp(type, ":jni") == 0)
                 args->verbosedll = TRUE;
 
-        } else if(strcmp(argv[i], "-Xnoasyncgc") == 0)
-            args->noasyncgc = TRUE;
+        } else if(strcmp(argv[i], "-Xasyncgc") == 0)
+            args->asyncgc = TRUE;
+
+        else if(strncmp(argv[i], "-ms", 3) == 0 ||
+                strncmp(argv[i], "-Xms", 4) == 0) {
+
+            char *value = argv[i] + (argv[i][1] == 'm' ? 3 : 4);
+            args->min_heap = parseMemValue(value);
 
-        else if(strncmp(argv[i], "-ms", 3) == 0 || strncmp(argv[i], "-Xms", 4) == 0) {
-            args->min_heap = parseMemValue(argv[i] + (argv[i][1] == 'm' ? 3 : 4));
             if(args->min_heap < MIN_HEAP) {
-                printf("Invalid minimum heap size: %s (min %dK)\n", argv[i], MIN_HEAP/KB);
+                printf("Invalid minimum heap size: %s (min %dK)\n", argv[i],
+                       MIN_HEAP/KB);
                 goto exit;
             }
 
-        } else if(strncmp(argv[i], "-mx", 3) == 0 || strncmp(argv[i], "-Xmx", 4) == 0) {
-            args->max_heap = parseMemValue(argv[i] + (argv[i][1] == 'm' ? 3 : 4));
+        } else if(strncmp(argv[i], "-mx", 3) == 0 ||
+                  strncmp(argv[i], "-Xmx", 4) == 0) {
+
+            char *value = argv[i] + (argv[i][1] == 'm' ? 3 : 4);
+            args->max_heap = parseMemValue(value);
+
             if(args->max_heap < MIN_HEAP) {
-                printf("Invalid maximum heap size: %s (min is %dK)\n", argv[i], MIN_HEAP/KB);
+                printf("Invalid maximum heap size: %s (min is %dK)\n", argv[i],
+                       MIN_HEAP/KB);
                 goto exit;
             }
 
-        } else if(strncmp(argv[i], "-ss", 3) == 0 || strncmp(argv[i], "-Xss", 4) == 0) {
-            args->java_stack = parseMemValue(argv[i] + (argv[i][1] == 's' ? 3 : 4));
+        } else if(strncmp(argv[i], "-ss", 3) == 0 ||
+                  strncmp(argv[i], "-Xss", 4) == 0) {
+
+            char *value = argv[i] + (argv[i][1] == 'm' ? 3 : 4);
+            args->java_stack = parseMemValue(value);
+
             if(args->java_stack < MIN_STACK) {
-                printf("Invalid Java stack size: %s (min is %dK)\n", argv[i], MIN_STACK/KB);
+                printf("Invalid Java stack size: %s (min is %dK)\n", argv[i],
+                       MIN_STACK/KB);
                 goto exit;
             }
 
         } else if(strncmp(argv[i], "-D", 2) == 0) {
-            char *pntr, *key = strcpy(sysMalloc(strlen(argv[i]+2) + 1), argv[i]+2);
+            char *key = strcpy(sysMalloc(strlen(argv[i] + 2) + 1), argv[i] + 2);
+            char *pntr;
+
             for(pntr = key; *pntr && (*pntr != '='); pntr++);
             if(*pntr)
                 *pntr++ = '\0';
             props[props_count].key = key;
             props[props_count++].value = pntr;
 
-        } else if(strcmp(argv[i], "-classpath") == 0 || strcmp(argv[i], "-cp") == 0) {
+        } else if(strcmp(argv[i], "-classpath") == 0 ||
+                  strcmp(argv[i], "-cp") == 0) {
 
-            if(i == (argc-1)) {
+            if(i == argc - 1) {
                 printf("%s : missing path list\n", argv[i]);
                 goto exit;
             }
@@ -258,16 +286,34 @@ int parseCommandLine(int argc, char *argv[], InitArgs *args) {
                code memory to zero */
             args->codemem = 0;
 
+        } else if(strcmp(argv[i], "-Xnoprofiling") == 0) {
+            args->profiling = FALSE;
+
+        } else if(strcmp(argv[i], "-Xnopatching") == 0) {
+            args->branch_patching = FALSE;
+
+        } else if(strcmp(argv[i], "-Xnopatchingdup") == 0) {
+            args->branch_patching_dup = FALSE;
+
+        } else if(strcmp(argv[i], "-Xnojoinblocks") == 0) {
+            args->join_blocks = FALSE;
+
+        } else if(strcmp(argv[i], "-Xcodestats") == 0) {
+            args->print_codestats = TRUE;
+
+        } else if(strncmp(argv[i], "-Xprofiling:", 12) == 0) {
+            args->profile_threshold = strtol(argv[i] + 12, NULL, 0);
+
         } else if(strncmp(argv[i], "-Xreplication:", 14) == 0) {
             char *pntr = argv[i] + 14;
 
             if(strcmp(pntr, "none") == 0)
-                args->replication = INT_MAX;
+                args->replication_threshold = INT_MAX;
             else
                 if(strcmp(pntr, "always") == 0)
-                    args->replication = 0;
+                    args->replication_threshold = 0;
                 else
-                    args->replication = strtol(pntr, NULL, 0);
+                    args->replication_threshold = strtol(pntr, NULL, 0);
 
         } else if(strncmp(argv[i], "-Xcodemem:", 10) == 0) {
             char *pntr = argv[i] + 10;
@@ -280,6 +326,12 @@ int parseCommandLine(int argc, char *argv[], InitArgs *args) {
             status = 0;
             goto exit;
 #endif
+        /* Compatibility options */
+        } else if(strcmp(argv[i], "-client") == 0 ||
+                  strcmp(argv[i], "-server") == 0 ||
+                  strncmp(argv[i], "-XX:PermSize=", 13) == 0 ||
+                  strncmp(argv[i], "-XX:MaxPermSize=", 16) == 0) {
+            /* Ignore */
         } else {
             printf("Unrecognised command line option: %s\n", argv[i]);
             break;
@@ -308,11 +360,8 @@ int main(int argc, char *argv[]) {
     args.main_stack_base = &array_class;
     initVM(&args);
 
-   if((system_loader = getSystemClassLoader()) == NULL) {
-        printf("Cannot create system class loader\n");
-        printException();
-        exitVM(1);
-    }
+   if((system_loader = getSystemClassLoader()) == NULL)
+        goto error;
 
     mainThreadSetContextClassLoader(system_loader);
 
@@ -320,18 +369,19 @@ int main(int argc, char *argv[]) {
         if(*cpntr == '.')
             *cpntr = '/';
 
-    if((main_class = findClassFromClassLoader(argv[class_arg], system_loader)) != NULL)
+    main_class = findClassFromClassLoader(argv[class_arg], system_loader);
+    if(main_class != NULL)
         initClass(main_class);
 
-    if(exceptionOccurred()) {
-        printException();
-        exitVM(1);
-    }
+    if(exceptionOccurred())
+        goto error;
 
-    mb = lookupMethod(main_class, SYMBOL(main), SYMBOL(_array_java_lang_String__V));
-    if(!mb || !(mb->access_flags & ACC_STATIC)) {
-        printf("Static method \"main\" not found in %s\n", argv[class_arg]);
-        exitVM(1);
+    mb = lookupMethod(main_class, SYMBOL(main),
+                                  SYMBOL(_array_java_lang_String__V));
+
+    if(mb == NULL || !(mb->access_flags & ACC_STATIC)) {
+        signalException(java_lang_NoSuchMethodError, "main");
+        goto error;
     }
 
     /* Create the String array holding the command line args */
@@ -339,7 +389,7 @@ int main(int argc, char *argv[]) {
     i = class_arg + 1;
     if((array_class = findArrayClass(SYMBOL(array_java_lang_String))) &&
            (array = allocArray(array_class, argc - i, sizeof(Object*))))  {
-        Object **args = (Object**)ARRAY_DATA(array) - i;
+        Object **args = ARRAY_DATA(array, Object*) - i;
 
         for(; i < argc; i++)
             if(!(args[i] = Cstr2String(argv[i])))
@@ -350,12 +400,16 @@ int main(int argc, char *argv[]) {
             executeStaticMethod(main_class, mb, array);
     }
 
+error:
     /* ExceptionOccurred returns the exception or NULL, which is OK
        for normal conditionals, but not here... */
     if((status = exceptionOccurred() ? 1 : 0))
-        printException();
+        uncaughtException();
 
     /* Wait for all but daemon threads to die */
     mainThreadWaitToExitVM();
     exitVM(status);
+
+   /* Keep the compiler happy */
+    return 0;
 }
