@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2003, 2004, 2005, 2006, 2007, 2008
+ * Copyright (C) 2003, 2004, 2005, 2006, 2007, 2008, 2009
  * Robert Lougher <rob@lougher.org.uk>.
  *
  * This file is part of JamVM.
@@ -19,10 +19,11 @@
  * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  */
 
-#include <stdarg.h>
 #include <inttypes.h>
+#include <stdarg.h>
 #include <limits.h>
 #include <stdio.h>
+#include <time.h>
 
 /* Configure options */
 #include "config.h"
@@ -251,6 +252,11 @@
 #define OPC_INVOKEVIRTUAL_QUICK         214
 #define OPC_INVOKENONVIRTUAL_QUICK      215
 #define OPC_INVOKESUPER_QUICK           216
+#define OPC_GETFIELD_QUICK_REF          217
+#define OPC_PUTFIELD_QUICK_REF          218
+#define OPC_GETSTATIC_QUICK_REF         219
+#define OPC_PUTSTATIC_QUICK_REF         220
+#define OPC_GETFIELD_THIS_REF           221
 #define OPC_INVOKEVIRTUAL_QUICK_W       226
 #define OPC_GETFIELD_QUICK_W            227
 #define OPC_PUTFIELD_QUICK_W            228
@@ -266,6 +272,7 @@
 #define OPC_INVOKEINTERFACE_QUICK       244
 #define OPC_ABSTRACT_METHOD_ERROR       245
 #define OPC_INLINE_REWRITER             246
+#define OPC_PROFILE_REWRITER            247
 
 #define CONSTANT_Utf8                   1
 #define CONSTANT_Integer                3
@@ -321,6 +328,18 @@
 #define CLASS_ARRAY             6
 #define CLASS_PRIM              7
 
+/* Internal primitive type numbers */
+
+#define PRIM_IDX_VOID           0
+#define PRIM_IDX_BOOLEAN        1
+#define PRIM_IDX_BYTE           2
+#define PRIM_IDX_CHAR           3
+#define PRIM_IDX_SHORT          4
+#define PRIM_IDX_INT            5
+#define PRIM_IDX_FLOAT          6
+#define PRIM_IDX_LONG           7
+#define PRIM_IDX_DOUBLE         8
+
 /* Class flags */
 
 #define CLASS_CLASS             1
@@ -420,22 +439,44 @@ typedef struct opcode_info {
     unsigned char cache_depth;
 } OpcodeInfo;
 
-typedef struct code_block {
+typedef struct profile_info ProfileInfo;
+
+typedef struct basic_block {
+    union {
+        struct {
+            int quickened;
+            ProfileInfo *profiled;
+        } profile;
+        struct {
+            char *addr;
+            struct basic_block *next;
+        } patch;
+    } u;
     int length;
     Instruction *start;
     OpcodeInfo *opcodes;
-} CodeBlock;
+    struct basic_block *prev;
+    struct basic_block *next;
+} BasicBlock;
 
 typedef struct quick_prepare_info {
-    CodeBlock block;
+    BasicBlock *block;
     Instruction *quickened;
     struct quick_prepare_info *next;
 } QuickPrepareInfo;
 
 typedef struct prepare_info {
-    CodeBlock block;
+    BasicBlock *block;
     Operand operand;
 } PrepareInfo;
+
+struct profile_info {
+    BasicBlock *block;
+    int profile_count;
+    const void *handler;
+    struct profile_info *next;
+    struct profile_info *prev;
+};
 #endif
 
 typedef Instruction *CodePntr;
@@ -477,6 +518,7 @@ typedef struct methodblock {
    MethodAnnotationData *annotations;
 #ifdef INLINING
    QuickPrepareInfo *quick_prepare_info;
+   ProfileInfo *profile_info;
 #endif
 } MethodBlock;
 
@@ -487,9 +529,17 @@ typedef struct fieldblock {
    char *signature;
    u2 access_flags;
    u2 constant;
-   uintptr_t static_value;
-   u4 offset;
    AnnotationData *annotations;
+   union {
+       union {
+           char data[8];
+           uintptr_t u;
+           long long l;
+           void *p;
+           int i;
+       } static_value; 
+       u4 offset;
+   } u;
 } FieldBlock;
 
 typedef struct itable_entry {
@@ -574,7 +624,7 @@ typedef struct prop {
 } Property;
 
 typedef struct InitArgs {
-    int noasyncgc;
+    int asyncgc;
     int verbosegc;
     int verbosedll;
     int verboseclass;
@@ -604,40 +654,49 @@ typedef struct InitArgs {
 
 #ifdef INLINING
     unsigned int codemem;
-    int replication;
+    int replication_threshold;
+    int profile_threshold;
+    int branch_patching_dup;
+    int branch_patching;
+    int print_codestats;
+    int join_blocks;
+    int profiling;
 #endif
 } InitArgs;
 
-#define CLASS_CB(classRef)              ((ClassBlock*)(classRef+1))
-#define INST_DATA(objectRef)            ((uintptr_t*)(objectRef+1))
+#define CLASS_CB(classRef)           ((ClassBlock*)(classRef+1))
 
-#define ARRAY_DATA(arrayRef)            ((void*)(((u4*)(arrayRef+1))+1))
-#define ARRAY_LEN(arrayRef)             *(u4*)(arrayRef+1)
+#define INST_DATA(obj, type, offset) *(type*)&((char*)obj)[offset]
+#define INST_BASE(obj, type)         ((type*)(obj+1))
 
-#define IS_CLASS(object)                (object->class && IS_CLASS_CLASS(CLASS_CB(object->class)))
+#define ARRAY_DATA(arrayRef, type)   ((type*)(((uintptr_t*)(arrayRef+1))+1)) 
+#define ARRAY_LEN(arrayRef)          *(uintptr_t*)(arrayRef+1)
 
-#define IS_INTERFACE(cb)                (cb->access_flags & ACC_INTERFACE)
-#define IS_SYNTHETIC(cb)                (cb->access_flags & ACC_SYNTHETIC)
-#define IS_ANNOTATION(cb)               (cb->access_flags & ACC_ANNOTATION)
-#define IS_ENUM(cb)                     (cb->access_flags & ACC_ENUM)
-#define IS_ARRAY(cb)                    (cb->state == CLASS_ARRAY)
-#define IS_PRIMITIVE(cb)                (cb->state >= CLASS_PRIM)
+#define IS_CLASS(object)             (object->class && IS_CLASS_CLASS( \
+                                                  CLASS_CB(object->class)))
 
-#define IS_FINALIZED(cb)                (cb->flags & FINALIZED)
-#define IS_REFERENCE(cb)		(cb->flags & REFERENCE)
-#define IS_SOFT_REFERENCE(cb)		(cb->flags & SOFT_REFERENCE)
-#define IS_WEAK_REFERENCE(cb)		(cb->flags & WEAK_REFERENCE)
-#define IS_PHANTOM_REFERENCE(cb)	(cb->flags & PHANTOM_REFERENCE)
-#define IS_CLASS_LOADER(cb)		(cb->flags & CLASS_LOADER)
-#define IS_CLASS_DUP(cb)		(cb->flags & CLASS_CLASH)
-#define IS_CLASS_CLASS(cb)		(cb->flags & CLASS_CLASS)
-#define IS_VMTHROWABLE(cb)		(cb->flags & VMTHROWABLE)
-#define IS_VMTHREAD(cb)			(cb->flags & VMTHREAD)
-#define IS_ANONYMOUS(cb)		(cb->flags & ANONYMOUS)
-#define IS_SPECIAL(cb)			(cb->flags & (REFERENCE | CLASS_LOADER | VMTHREAD))
-
-#define IS_MEMBER(cb)			cb->declaring_class
-#define IS_LOCAL(cb)			(cb->enclosing_method && !IS_ANONYMOUS(cb))
+#define IS_INTERFACE(cb)             (cb->access_flags & ACC_INTERFACE)
+#define IS_SYNTHETIC(cb)             (cb->access_flags & ACC_SYNTHETIC)
+#define IS_ANNOTATION(cb)            (cb->access_flags & ACC_ANNOTATION)
+#define IS_ENUM(cb)                  (cb->access_flags & ACC_ENUM)
+#define IS_ARRAY(cb)                 (cb->state == CLASS_ARRAY)
+#define IS_PRIMITIVE(cb)             (cb->state >= CLASS_PRIM)
+ 
+#define IS_FINALIZED(cb)             (cb->flags & FINALIZED)
+#define IS_REFERENCE(cb)             (cb->flags & REFERENCE)
+#define IS_SOFT_REFERENCE(cb)        (cb->flags & SOFT_REFERENCE)
+#define IS_WEAK_REFERENCE(cb)        (cb->flags & WEAK_REFERENCE)
+#define IS_PHANTOM_REFERENCE(cb)     (cb->flags & PHANTOM_REFERENCE)
+#define IS_CLASS_LOADER(cb)          (cb->flags & CLASS_LOADER)
+#define IS_CLASS_DUP(cb)             (cb->flags & CLASS_CLASH)
+#define IS_CLASS_CLASS(cb)           (cb->flags & CLASS_CLASS)
+#define IS_VMTHROWABLE(cb)           (cb->flags & VMTHROWABLE)
+#define IS_VMTHREAD(cb)              (cb->flags & VMTHREAD)
+#define IS_ANONYMOUS(cb)             (cb->flags & ANONYMOUS)
+#define IS_SPECIAL(cb)               (cb->flags & (REFERENCE | CLASS_LOADER | \
+                                                  VMTHREAD))
+#define IS_MEMBER(cb)                cb->declaring_class
+#define IS_LOCAL(cb)                 (cb->enclosing_method && !IS_ANONYMOUS(cb))
 
 /* Macros for accessing constant pool entries */
 
@@ -671,7 +730,7 @@ typedef struct InitArgs {
 
 /* default minimum size of object heap */
 #ifndef DEFAULT_MIN_HEAP
-#define DEFAULT_MIN_HEAP 2*MB
+#define DEFAULT_MIN_HEAP 16*MB
 #endif
 
 /* default maximum size of object heap */
@@ -680,7 +739,7 @@ typedef struct InitArgs {
 #endif
 
 /* default size of the Java stack */
-#define DEFAULT_STACK 64*KB
+#define DEFAULT_STACK 256*KB
 
 /* size of emergency area - big enough to create
    a StackOverflow exception */
@@ -720,14 +779,14 @@ extern void *gcMemMalloc(int n);
 extern void gcMemFree(void *ptr);
 extern void *gcMemRealloc(void *ptr, int n);
 
-extern void registerStaticObjectRef(Object **ob);
-extern void registerStaticObjectRefLocked(Object **ob);
+extern void registerStaticObjectRef(Object **ref);
+extern void registerStaticObjectRefLocked(Object **ref, Object *obj);
 
 #define registerStaticClassRef(ref) \
     registerStaticObjectRef(ref);
 
-#define registerStaticClassRefLocked(ref) \
-    registerStaticObjectRefLocked(ref);
+#define registerStaticClassRefLocked(ref, class) \
+    registerStaticObjectRefLocked(ref, class);
 
 extern void gcPendingFree(void *addr);
 
@@ -739,7 +798,8 @@ extern int isMarked(Object *ob);
 
 extern Class *java_lang_Class;
 
-extern Class *defineClass(char *classname, char *data, int offset, int len, Object *class_loader);
+extern Class *defineClass(char *classname, char *data, int offset, int len,
+                          Object *class_loader);
 extern void linkClass(Class *class);
 extern Class *initClass(Class *class);
 extern Class *findSystemClass(char *);
@@ -756,12 +816,12 @@ extern int bootClassPathSize();
 extern Object *bootClassPathResource(char *filename, int index);
 
 #define findArrayClassFromClass(name, class) \
-                    findArrayClassFromClassLoader(name, CLASS_CB(class)->class_loader)
+             findArrayClassFromClassLoader(name, CLASS_CB(class)->class_loader)
 #define findArrayClass(name) findArrayClassFromClassLoader(name, NULL)
 
 extern Class *findClassFromClassLoader(char *, Object *);
 #define findClassFromClass(name, class) \
-                    findClassFromClassLoader(name, CLASS_CB(class)->class_loader)
+             findClassFromClassLoader(name, CLASS_CB(class)->class_loader)
 
 extern void freeClassData(Class *class);
 extern void freeClassLoaderData(Object *class_loader);
@@ -776,6 +836,9 @@ extern void threadLoaderClasses(Object *class_loader);
 extern void newLibraryUnloader(Object *class_loader, void *entry);
 extern void initialiseClass(InitArgs *args);
 
+extern Object *bootPackage(char *package_name);
+extern Object *bootPackages();
+
 /* resolve */
 
 extern FieldBlock *findField(Class *, char *, char *);
@@ -799,8 +862,10 @@ extern char arrayStoreCheck(Class *class, Class *test);
 /* execute */
 
 extern void *executeMethodArgs(Object *ob, Class *class, MethodBlock *mb, ...);
-extern void *executeMethodVaList(Object *ob, Class *class, MethodBlock *mb, va_list args);
-extern void *executeMethodList(Object *ob, Class *class, MethodBlock *mb, u8 *args);
+extern void *executeMethodVaList(Object *ob, Class *class, MethodBlock *mb,
+                                  va_list args);
+extern void *executeMethodList(Object *ob, Class *class, MethodBlock *mb,
+                               u8 *args);
 
 #define executeMethod(ob, mb, args...) \
     executeMethodArgs(ob, ob->class, mb, ##args)
@@ -811,9 +876,12 @@ extern void *executeMethodList(Object *ob, Class *class, MethodBlock *mb, u8 *ar
 /* excep */
 
 extern Object *exceptionOccurred();
-extern void signalChainedExceptionEnum(int excep_enum, char *excep_mess, Object *cause);
-extern void signalChainedExceptionName(char *excep_name, char *excep_mess, Object *cause);
-extern void signalChainedExceptionClass(Class *excep_class, char *excep_mess, Object *cause);
+extern void signalChainedExceptionEnum(int excep_enum, char *excep_mess,
+                                       Object *cause);
+extern void signalChainedExceptionName(char *excep_name, char *excep_mess,
+                                       Object *cause);
+extern void signalChainedExceptionClass(Class *excep_class, char *excep_mess,
+                                        Object *cause);
 extern void setException(Object *excep);
 extern void clearException();
 extern void printException();
@@ -824,7 +892,7 @@ extern int mapPC2LineNo(MethodBlock *mb, CodePntr pc_pntr);
 extern void markVMThrowable(Object *vmthrwble, int mark, int mark_soft_refs);
 extern void initialiseException();
 
-#define exceptionOccured0(ee) \
+#define exceptionOccurred0(ee) \
     ee->exception
 
 #define signalException(excep_name, excep_mess) \
@@ -842,6 +910,7 @@ extern void initialiseException();
 /* interp */
 
 extern uintptr_t *executeJava();
+extern void shutdownInterpreter();
 extern void initialiseInterpreter(InitArgs *args);
 
 /* String */
@@ -891,7 +960,8 @@ extern char *getDllPath();
 extern char *getBootDllPath();
 extern char *getDllName(char *name);
 extern void initialiseDll(InitArgs *args);
-extern uintptr_t *resolveNativeWrapper(Class *class, MethodBlock *mb, uintptr_t *ostack);
+extern uintptr_t *resolveNativeWrapper(Class *class, MethodBlock *mb,
+                                       uintptr_t *ostack);
 extern void unloaderUnloadDll(uintptr_t entry);
 extern void unloadClassLoaderDlls(Object *loader);
 extern void threadLiveClassLoaderDlls();
@@ -915,6 +985,7 @@ extern ExecEnv *getExecEnv();
 extern void createJavaThread(Object *jThread, long long stack_size);
 extern void mainThreadSetContextClassLoader(Object *loader);
 extern void mainThreadWaitToExitVM();
+extern void uncaughtException();
 extern void exitVM(int status);
 extern void scanThreads();
 
@@ -922,39 +993,6 @@ extern void scanThreads();
 
 extern void initialiseMonitor();
 
-/* reflect */
-
-extern Object *getClassConstructors(Class *class, int public);
-extern Object *getClassMethods(Class *class, int public);
-extern Object *getClassFields(Class *class, int public);
-extern Object *getClassInterfaces(Class *class);
-extern Object *getClassClasses(Class *class, int public);
-extern Class *getDeclaringClass(Class *class);
-extern Class *getEnclosingClass(Class *class);
-extern Object *getEnclosingMethodObject(Class *class);
-extern Object *getEnclosingConstructorObject(Class *class);
-extern Object *getClassAnnotations(Class *class);
-extern Object *getFieldAnnotations(FieldBlock *fb);
-extern Object *getMethodAnnotations(MethodBlock *mb);
-extern Object *getMethodParameterAnnotations(MethodBlock *mb);
-extern Object *getMethodDefaultValue(MethodBlock *mb);
-
-extern Object *getReflectReturnObject(Class *type, uintptr_t *pntr);
-extern uintptr_t *widenPrimitiveValue(int src_idx, int dest_idx, uintptr_t *src, uintptr_t *dest);
-extern uintptr_t *unwrapAndWidenObject(Class *type, Object *arg, uintptr_t *pntr);
-extern Object *invoke(Object *ob, MethodBlock *mb, Object *arg_array, Object *param_types,
-                      int check_access);
-
-extern MethodBlock *mbFromReflectObject(Object *reflect_ob);
-extern FieldBlock *fbFromReflectObject(Object *reflect_ob);
-
-extern Object *createReflectConstructorObject(MethodBlock *mb);
-extern Object *createReflectMethodObject(MethodBlock *mb);
-extern Object *createReflectFieldObject(FieldBlock *fb);
-extern Class *getReflectMethodClass();
-
-#define getPrimTypeIndex(cb) (cb->state - CLASS_PRIM)
-
 /* jni */
 
 extern int initJNILrefs();
@@ -968,6 +1006,7 @@ extern void initialiseProperties(InitArgs *args);
 extern void addCommandLineProperties(Object *properties);
 extern void addDefaultProperties(Object *properties);
 extern char *getCommandLineProperty(char *key);
+extern char *getCwd();
 
 /* access */
 
@@ -991,6 +1030,10 @@ extern unsigned long parseMemValue(char *str);
 extern void initVM(InitArgs *args);
 extern int VMInitialising();
 
+/* shutdown */
+
+extern void shutdownVM(int status);
+
 /* hooks */
 
 extern void initialiseHooks(InitArgs *args);
@@ -1004,7 +1047,16 @@ extern void jamvm_exit(int status);
 extern void freeMethodInlinedInfo(MethodBlock *mb);
 extern int  initialiseInlining(InitArgs *args);
 extern void showRelocatability();
+extern void shutdownInlining();
 
 /* symbol */
+
 extern void initialiseSymbol();
 
+/* time */
+
+extern void getTimeoutAbsolute(struct timespec *ts, long long millis,
+                        long long nanos);
+extern void getTimeoutRelative(struct timespec *ts, long long millis,
+                        long long nanos);
+
