@@ -19,6 +19,7 @@
  * Foundation, 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301, USA.
  */
 
+#include <errno.h>
 #include <string.h>
 #include <stdlib.h>
 #include <unistd.h>
@@ -132,13 +133,30 @@ void setEndianProperty(Object *properties) {
 #endif
 }
 
+char *getCwd() {
+    char *cwd = NULL;
+    int size = 256;
+
+    for(;;) {
+        cwd = sysRealloc(cwd, size);
+
+        if(getcwd(cwd, size) == NULL)
+            if(errno == ERANGE)
+               size *= 2;
+            else {
+               perror("Couldn't get cwd");
+               exitVM(1);
+            }
+        else
+            return cwd;
+    }
+}
+    
 void setUserDirProperty(Object *properties) {
-    char *cwd = getcwd(NULL, 0);
+    char *cwd = getCwd();
 
     setProperty(properties, "user.dir", cwd);
-
-    if(cwd != NULL)
-        sysFree(cwd);
+    sysFree(cwd);
 }
 
 void setOSProperties(Object *properties) {
@@ -158,9 +176,13 @@ char *getJavaHome() {
 void addDefaultProperties(Object *properties) {
     setProperty(properties, "java.vm.name", "JamVM");
     setProperty(properties, "java.vm.version", VERSION);
-    setProperty(properties, "java.runtime.version", VERSION);
     setProperty(properties, "java.vm.vendor", "Robert Lougher");
-    setProperty(properties, "java.vm.vendor.url", "http://jamvm.sourceforge.net");
+    setProperty(properties, "java.vm.vendor.url", "http://jamvm.org");
+    setProperty(properties, "java.vm.specification.version", "1.0");
+    setProperty(properties, "java.vm.specification.vendor", "Sun Microsystems, Inc.");
+    setProperty(properties, "java.vm.specification.name", "Java Virtual Machine Specification");
+
+    setProperty(properties, "java.runtime.version", VERSION);
     setProperty(properties, "java.version", JAVA_COMPAT_VERSION);
     setProperty(properties, "java.vendor", "GNU Classpath");
     setProperty(properties, "java.vendor.url", "http://www.classpath.org");
@@ -168,9 +190,6 @@ void addDefaultProperties(Object *properties) {
     setProperty(properties, "java.specification.version", "1.5");
     setProperty(properties, "java.specification.vendor", "Sun Microsystems, Inc.");
     setProperty(properties, "java.specification.name", "Java Platform API Specification");
-    setProperty(properties, "java.vm.specification.version", "1.0");
-    setProperty(properties, "java.vm.specification.vendor", "Sun Microsystems, Inc.");
-    setProperty(properties, "java.vm.specification.name", "Java Virtual Machine Specification");
     setProperty(properties, "java.class.version", "48.0");
     setProperty(properties, "java.class.path", getClassPath());
     setProperty(properties, "sun.boot.class.path", getBootClassPath());
