@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2003, 2004, 2005, 2006, 2007, 2008
+ * Copyright (C) 2003, 2004, 2005, 2006, 2007, 2008, 2009
  * Robert Lougher <rob@lougher.org.uk>.
  *
  * This file is part of JamVM.
@@ -32,7 +32,7 @@
                   ((hash1 == hash2) && stringComp(ptr1, ptr2))
 #define PREPARE(ptr) ptr
 #define SCAVENGE(ptr) FALSE
-#define FOUND(ptr) ptr
+#define FOUND(ptr1, ptr2) ptr2
 
 static Class *string_class;
 static int count_offset; 
@@ -42,10 +42,10 @@ static int offset_offset;
 static HashTable hash_table;
 
 int stringHash(Object *ptr) {
-    Object *array = (Object*)INST_DATA(ptr)[value_offset]; 
-    int len = INST_DATA(ptr)[count_offset];
-    int offset = INST_DATA(ptr)[offset_offset];
-    unsigned short *dpntr = ((unsigned short *)ARRAY_DATA(array))+offset;
+    int len = INST_DATA(ptr, int, count_offset);
+    int offset = INST_DATA(ptr, int, offset_offset);
+    Object *array = INST_DATA(ptr, Object*, value_offset); 
+    unsigned short *dpntr = ARRAY_DATA(array, unsigned short) + offset;
     int hash = 0;
 
     for(; len > 0; len--)
@@ -55,16 +55,16 @@ int stringHash(Object *ptr) {
 }
 
 int stringComp(Object *ptr, Object *ptr2) {
-    int len = INST_DATA(ptr)[count_offset];
-    int len2 = INST_DATA(ptr2)[count_offset];
+    int len = INST_DATA(ptr, int, count_offset);
+    int len2 = INST_DATA(ptr2, int, count_offset);
 
     if(len == len2) {
-        Object *array = (Object*)INST_DATA(ptr)[value_offset];
-        Object *array2 = (Object*)INST_DATA(ptr2)[value_offset];
-        int offset = INST_DATA(ptr)[offset_offset];
-        int offset2 = INST_DATA(ptr2)[offset_offset];
-        unsigned short *src = ((unsigned short *)ARRAY_DATA(array))+offset;
-        unsigned short *dst = ((unsigned short *)ARRAY_DATA(array2))+offset2;
+        Object *array = INST_DATA(ptr, Object*, value_offset);
+        Object *array2 = INST_DATA(ptr2, Object*, value_offset);
+        int offset = INST_DATA(ptr, int, offset_offset);
+        int offset2 = INST_DATA(ptr2, int, offset_offset);
+        unsigned short *src = ARRAY_DATA(array, unsigned short) + offset;
+        unsigned short *dst = ARRAY_DATA(array2, unsigned short) + offset2;
 
         for(; (len > 0) && (*src++ == *dst++); len--);
 
@@ -85,11 +85,11 @@ Object *createString(char *utf8) {
        (ob = allocObject(string_class)) == NULL)
         return NULL;
 
-    data = ARRAY_DATA(array);
+    data = ARRAY_DATA(array, unsigned short);
     convertUtf8(utf8, data);
 
-    INST_DATA(ob)[count_offset] = len; 
-    INST_DATA(ob)[value_offset] = (uintptr_t)array; 
+    INST_DATA(ob, int, count_offset) = len; 
+    INST_DATA(ob, Object*, value_offset) = array; 
 
     return ob;
 }
@@ -138,9 +138,9 @@ void threadInternedStrings() {
 }
 
 char *String2Buff0(Object *string, char *buff, int len) {
-    int offset = INST_DATA(string)[offset_offset];
-    Object *array =(Object *)(INST_DATA(string)[value_offset]);
-    unsigned short *str = ((unsigned short *)ARRAY_DATA(array))+offset;
+    int offset = INST_DATA(string, int, offset_offset);
+    Object *array = INST_DATA(string, Object*, value_offset);
+    unsigned short *str = ARRAY_DATA(array, unsigned short) + offset;
     char *pntr;
 
     for(pntr = buff; len > 0; len--)
@@ -151,21 +151,21 @@ char *String2Buff0(Object *string, char *buff, int len) {
 }
 
 char *String2Buff(Object *string, char *buff, int buff_len) {
-    int str_len = INST_DATA(string)[count_offset];
+    int str_len = INST_DATA(string, int, count_offset);
     int len = buff_len-1 < str_len ? buff_len-1 : str_len;
 
     return String2Buff0(string, buff, len);
 }
 
 char *String2Cstr(Object *string) {
-    int len = INST_DATA(string)[count_offset];
+    int len = INST_DATA(string, int, count_offset);
     char *buff = (char *)sysMalloc(len + 1);
 
     return String2Buff0(string, buff, len);
 }
 
 void initialiseString() {
-    FieldBlock *count = NULL, *value, *offset;
+    FieldBlock *count = NULL, *value = NULL, *offset = NULL;
 
     string_class = findSystemClass0(SYMBOL(java_lang_String));
     registerStaticClassRef(&string_class);
@@ -182,9 +182,9 @@ void initialiseString() {
         exitVM(1);
     }
 
-    count_offset = count->offset;
-    value_offset = value->offset;
-    offset_offset = offset->offset;
+    count_offset = count->u.offset;
+    value_offset = value->u.offset;
+    offset_offset = offset->u.offset;
 
     /* Init hash table and create lock */
     initHashTable(hash_table, HASHTABSZE, TRUE);
@@ -198,28 +198,28 @@ Object *createStringFromUnicode(unsigned short *unicode, int len) {
     Object *ob = allocObject(string_class);
 
     if(array != NULL && ob != NULL) {
-        unsigned short *data = ARRAY_DATA(array);
-        memcpy(data, unicode, len*sizeof(unsigned short));
+        unsigned short *data = ARRAY_DATA(array, unsigned short);
+        memcpy(data, unicode, len * sizeof(unsigned short));
 
-        INST_DATA(ob)[count_offset] = len; 
-        INST_DATA(ob)[value_offset] = (uintptr_t)array; 
+        INST_DATA(ob, int, count_offset) = len; 
+        INST_DATA(ob, Object*, value_offset) = array; 
         return ob;
     }
     return NULL;
 }
 
 Object *getStringCharsArray(Object *string) {
-    return (Object*)INST_DATA(string)[value_offset];
+    return INST_DATA(string, Object*, value_offset);
 }
 
 unsigned short *getStringChars(Object *string) {
-    Object *array = (Object*)INST_DATA(string)[value_offset];
-    int offset = INST_DATA(string)[offset_offset];
-    return ((unsigned short*)ARRAY_DATA(array))+offset;
+    Object *array = INST_DATA(string, Object*, value_offset);
+    int offset = INST_DATA(string, int, offset_offset);
+    return ARRAY_DATA(array, unsigned short) + offset;
 }
 
 int getStringLen(Object *string) {
-    return INST_DATA(string)[count_offset];
+    return INST_DATA(string, int, count_offset);
 }
 
 int getStringUtf8Len(Object *string) {
