@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2003, 2004, 2005, 2006, 2007, 2008
+ * Copyright (C) 2003, 2004, 2005, 2006, 2007, 2008, 2009
  * Robert Lougher <rob@lougher.org.uk>.
  *
  * This file is part of JamVM.
@@ -22,6 +22,12 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <string.h>
+
+/* Required on OpenSolaris to get standard conforming sigwait. */
+#ifndef _POSIX_PTHREAD_SEMANTICS
+#define _POSIX_PTHREAD_SEMANTICS
+#endif
+
 #include <signal.h>
 #include <sched.h>
 
@@ -56,8 +62,14 @@ static pthread_cond_t exit_cv;
 /* Monitor for sleeping threads to do a timed-wait against */
 static Monitor sleep_mon;
 
-/* Thread specific key holding thread's Thread pntr */
-static pthread_key_t threadKey;
+/* If supported, use thread local storage to store the
+   thread's Thread pntr.  If not, use a pthread thread
+   specific key to hold it */
+#ifdef HAVE_TLS
+static __thread Thread *self = NULL;
+#else
+static pthread_key_t self;
+#endif
 
 /* Attributes for spawned threads */
 static pthread_attr_t attributes;
@@ -141,6 +153,24 @@ int threadIsAlive(Thread *thread) {
     return thread->state != 0;
 }
 
+void threadSleep(Thread *self, long long ms, int ns) {
+
+   /* A sleep of zero should just yield, but a wait
+      of zero is "forever".  Therefore wait for a tiny
+      amount -- this should yield and we also get the
+      interrupted checks. */
+    if(ms == 0 &&  ns == 0)
+        ns = 1;
+
+    monitorLock(&sleep_mon, self);
+    monitorWait(&sleep_mon, self, ms, ns);
+    monitorUnlock(&sleep_mon, self);
+}
+
+void threadYield(Thread *thread) {
+    sched_yield();
+}
+
 int threadInterrupted(Thread *thread) {
     int r = thread->interrupted;
     thread->interrupted = FALSE;
@@ -151,32 +181,14 @@ int threadIsInterrupted(Thread *thread) {
     return thread->interrupted;
 }
 
-void threadSleep(Thread *thread, long long ms, int ns) {
-
-   /* A sleep of zero should just yield, but a wait
-      of zero is "forever".  Therefore wait for a tiny
-      amount -- this should yield and we also get the
-      interrupted checks. */
-    if(ms == 0 &&  ns == 0)
-        ns = 1;
-
-    monitorLock(&sleep_mon, thread);
-    monitorWait(&sleep_mon, thread, ms, ns);
-    monitorUnlock(&sleep_mon, thread);
-}
-
-void threadYield(Thread *thread) {
-    sched_yield();
-}
-
 void threadInterrupt(Thread *thread) {
-    Monitor *mon;
     Thread *self = threadSelf();
+    Monitor *mon;
 
-    /* monitorWait sets wait_mon _before_ checking interrupted status.  Therefore,
-       if wait_mon is null, interrupted status will be noticed.  This guards
-       against a race-condition leading to an interrupt being missed.  The memory
-       barrier ensures correct ordering on SMP systems. */
+    /* MonitorWait sets wait_mon _before_ checking interrupted status.
+       Therefore, if wait_mon is null, interrupted status will be noticed.
+       This guards against a race-condition leading to an interrupt being
+       missed.  The memory barrier ensures correct ordering on SMP systems. */
     thread->interrupted = TRUE;
     MBARRIER();
 
@@ -184,9 +196,10 @@ void threadInterrupt(Thread *thread) {
         int locked;
         thread->interrupting = TRUE;
 
-        /* Ensure the thread has actually entered the wait
-           (else the signal will be lost) */
-
+        /* The thread is waiting on a monitor, but it may not have
+           entered the wait (in which case the signal will be lost).
+           Loop until we can get ownership (i.e. the thread has released
+           it on waiting) */
         while(!(locked = !pthread_mutex_trylock(&mon->lock)) && mon->owner == NULL)
             sched_yield();
 
@@ -196,6 +209,9 @@ void threadInterrupt(Thread *thread) {
             pthread_mutex_unlock(&mon->lock);
     }
 
+    /* Thread may still be parked */
+    threadUnpark(thread);
+
     /* Handle the case where the thread is blocked in a system call.
        This will knock it out with an EINTR.  The suspend signal
        handler will just return (as in the user doing a kill), and do
@@ -213,6 +229,88 @@ void threadInterrupt(Thread *thread) {
     fastEnableSuspend(self);
 }
 
+void threadPark(Thread *self, int absolute, long long time) {
+    /* If we have a permit use it and return immediately.
+       No locking as we're the only one that can change the
+       state at this point */
+    if(self->park_state == PARK_PERMIT) {
+        self->park_state = PARK_RUNNING;
+        MBARRIER();
+        return;
+    } 
+
+    /* Spin until we can get the park lock.  This avoids having
+       to disable suspension around pthread_mutex_lock */
+    while(pthread_mutex_trylock(&self->park_lock))
+        sched_yield();
+
+    /* A thread may have given us a permit while we were waiting
+       for the lock or we may be running.  Reduce the state by
+       one (PERMIT -> RUNNING, RUNNING -> BLOCKED) and wait if
+       we're now blocked */
+
+    if(--self->park_state == PARK_BLOCKED) {
+        /* Really must disable suspension now as we're
+           going to sleep */
+        disableSuspend(self);
+
+        if(time) {
+            struct timespec ts;
+
+            if(absolute)
+                getTimeoutAbsolute(&ts, time, 0);
+            else
+                getTimeoutRelative(&ts, 0, time);
+
+            self->state = TIMED_WAITING;
+            pthread_cond_timedwait(&self->park_cv, &self->park_lock, &ts);
+
+            /* On Linux/i386 systems using LinuxThreads, pthread_cond_timedwait
+               is implemented using sigjmp/longjmp.  This resets the fpu
+               control word back to 64-bit precision.  The macro is empty for
+               sane platforms. */ 
+
+            FPU_HACK;
+        } else {
+            self->state = WAITING;
+            pthread_cond_wait(&self->park_cv, &self->park_lock);
+        }
+        
+        /* If we were unparked park_state will have been updated,
+           but not if the wait timed out.  Only update if it's
+           blocked, to avoid losing a possible permit */
+        if(self->park_state == PARK_BLOCKED)
+            self->park_state = PARK_RUNNING;
+
+        self->state = RUNNING;
+
+        enableSuspend(self);
+    }
+
+    pthread_mutex_unlock(&self->park_lock);
+}
+
+void threadUnpark(Thread *thread) {
+    if(thread->park_state != PARK_PERMIT) {
+
+        /* Spin until we can get the park lock.  This avoids having
+           to disable suspension around pthread_mutex_lock */
+        while(pthread_mutex_trylock(&thread->park_lock))
+            sched_yield();
+
+        /* If another thread has given a permit while we were
+           waiting for the lock do nothing.  Else increase the
+           state by one (BLOCKED -> RUNNING, RUNNING -> PERMIT)
+           and if the thread was blocked signal it */
+
+        if(thread->park_state != PARK_PERMIT &&
+                  thread->park_state++ == PARK_BLOCKED)
+            pthread_cond_signal(&thread->park_cv);
+
+        pthread_mutex_unlock(&thread->park_lock);
+    }
+}
+
 void *getStackTop(Thread *thread) {
     return thread->stack_top;
 }
@@ -221,17 +319,29 @@ void *getStackBase(Thread *thread) {
     return thread->stack_base;
 }
 
-Thread *threadSelf0(Object *vmThread) {
-    return (Thread*)(INST_DATA(vmThread)[vmData_offset]);
+Thread *vmThread2Thread(Object *vmThread) {
+    return INST_DATA(vmThread, Thread*, vmData_offset);
+}
+
+Thread *jThread2Thread(Object *jThread) {
+    Object *vmthread = INST_DATA(jThread, Object*, vmthread_offset);
+    return vmthread == NULL ? NULL : vmThread2Thread(vmthread);
 }
 
 Thread *threadSelf() {
-    return (Thread*)pthread_getspecific(threadKey);
+#ifdef HAVE_TLS
+    return self;
+#else
+    return (Thread*)pthread_getspecific(self);
+#endif
 }
 
 void setThreadSelf(Thread *thread) {
-    pthread_setspecific(threadKey, thread);
-    pthread_cond_init(&thread->wait_cv, NULL);
+#ifdef HAVE_TLS
+    self = thread;
+#else
+    pthread_setspecific(self, thread);
+#endif
 }
 
 ExecEnv *getExecEnv() {
@@ -281,14 +391,16 @@ void resetPeakThreadsCount() {
 }
 
 void initialiseJavaStack(ExecEnv *ee) {
-   int stack_size = ee->stack_size ?
-           (ee->stack_size > MIN_STACK ? ee->stack_size : MIN_STACK) : dflt_stack_size;
+   int stack_size = ee->stack_size
+          ? (ee->stack_size > MIN_STACK ? ee->stack_size : MIN_STACK)
+          : dflt_stack_size;
    char *stack = sysMalloc(stack_size);
-   MethodBlock *mb = (MethodBlock *) stack;
-   Frame *top = (Frame *) (mb+1);
+   MethodBlock *mb = (MethodBlock *)stack;
+   Frame *top = (Frame *)(mb + 1);
 
-   top->ostack = (uintptr_t*)(top+1);
+   top->ostack = (uintptr_t*)(top + 1);
    top->lvars = (uintptr_t*)top;
+   mb->args_count = 0;
    mb->max_stack = 0;
    top->prev = NULL;
    top->mb = mb;
@@ -300,18 +412,18 @@ void initialiseJavaStack(ExecEnv *ee) {
 }
 
 long long javaThreadId(Thread *thread) {
-    return *(long long*)&(INST_DATA(thread->ee->thread)[threadId_offset]);
+    return INST_DATA(thread->ee->thread, long long, threadId_offset);
 }
 
 Thread *findHashedThread(Thread *thread, long long id) {
 
-#define FOUND(ptr) ptr
 #define DELETED ((void*)-1)
+#define FOUND(ptr1, ptr2) ptr2
 #define PREPARE(thread_id) thread
 #define SCAVENGE(ptr) ptr == DELETED
 #define HASH(thread_id) (int)thread_id
 #define COMPARE(thread_id, ptr, hash1, hash2) ptr != DELETED && \
-                                              (hash1 == hash2 && thread_id == javaThreadId(ptr))
+                             (hash1 == hash2 && thread_id == javaThreadId(ptr))
 
     Thread *ptr;
 
@@ -348,8 +460,8 @@ Object *initJavaThread(Thread *thread, char is_daemon, char *name) {
         return NULL;
 
     thread->ee->thread = jlthread;
-    INST_DATA(vmthread)[vmData_offset] = (uintptr_t)thread;
-    INST_DATA(vmthread)[thread_offset] = (uintptr_t)jlthread;
+    INST_DATA(vmthread, Thread*, vmData_offset) = thread;
+    INST_DATA(vmthread, Object*, thread_offset) = jlthread;
 
     /* Create the string for the thread name.  If null is specified
        the initialiser method will generate a name of the form Thread-X */
@@ -358,7 +470,11 @@ Object *initJavaThread(Thread *thread, char is_daemon, char *name) {
 
     /* Call the initialiser method -- this is for use by threads
        created or attached by the VM "outside" of Java */
-    executeMethod(jlthread, init_mb, vmthread, thread_name, NORM_PRIORITY, is_daemon);
+    executeMethod(jlthread, init_mb, vmthread, thread_name, NORM_PRIORITY,
+                  is_daemon);
+
+    if(exceptionOccurred())
+        return NULL;
 
     /* Add thread to thread ID map hash table. */
     addThreadToHash(thread);
@@ -373,6 +489,13 @@ void initThread(Thread *thread, char is_daemon, void *stack_base) {
     initialiseJavaStack(thread->ee);
     setThreadSelf(thread);
 
+    /* Initialise wait condvar (the condvar is per-thread,
+       not per-monitor) */
+    pthread_cond_init(&thread->wait_cv, NULL);
+    thread->park_state = PARK_RUNNING;
+    pthread_cond_init(&thread->park_cv, NULL);
+    pthread_mutex_init(&thread->park_lock, NULL);
+
     /* Record the thread's stack base */
     thread->stack_base = stack_base;
 
@@ -407,8 +530,9 @@ void initThread(Thread *thread, char is_daemon, void *stack_base) {
     if(!is_daemon)
         non_daemon_thrds++;
 
-    /* Get a thread ID (used in thin locking) and
-       record we're now running */
+    /* Get a thread ID (used in thin locking -- done here
+       as the thread lock is held) and record we're now
+       running */
     thread->id = genThreadID();
     thread->state = RUNNING;
 
@@ -416,7 +540,9 @@ void initThread(Thread *thread, char is_daemon, void *stack_base) {
     pthread_mutex_unlock(&lock);
 }
 
-Thread *attachThread(char *name, char is_daemon, void *stack_base, Thread *thread, Object *group) {
+Thread *attachThread(char *name, char is_daemon, void *stack_base,
+                     Thread *thread, Object *group) {
+
     Object *java_thread;
 
     /* Create the ExecEnv for the thread */
@@ -435,7 +561,7 @@ Thread *attachThread(char *name, char is_daemon, void *stack_base, Thread *threa
         return NULL;
 
     /* Initialiser doesn't handle the thread group */
-    INST_DATA(java_thread)[group_offset] = (uintptr_t)group;
+    INST_DATA(java_thread, Object*, group_offset) = group;
     executeMethod(group, addThread_mb, java_thread);
 
     /* We're now attached to the VM...*/
@@ -443,39 +569,58 @@ Thread *attachThread(char *name, char is_daemon, void *stack_base, Thread *threa
     return thread;
 }
 
-void detachThread(Thread *thread) {
-    Object *group, *excep;
+/* Call uncaughtException on the thread's exception handler, or the
+   thread's group if this is unset */
+
+void uncaughtException() {
+    Thread *thread = threadSelf();
     ExecEnv *ee = thread->ee;
     Object *jThread = ee->thread;
-    Object *vmthread = (Object*)INST_DATA(jThread)[vmthread_offset];
+    Object *excep = exceptionOccurred();
+    Object *group = INST_DATA(jThread, Object*, group_offset);
+    FieldBlock *fb = findField(thread_class,
+                        SYMBOL(exceptionHandler),
+                        SYMBOL(sig_java_lang_Thread_UncaughtExceptionHandler));
+    Object *thread_handler = fb == NULL ? NULL :
+                                  INST_DATA(jThread, Object*, fb->u.offset);
+    Object *handler = thread_handler == NULL ? group : thread_handler;
+    MethodBlock *uncaught_mb = lookupMethod(handler->class,
+                              SYMBOL(uncaughtException),
+                              SYMBOL(_java_lang_Thread_java_lang_Throwable__V));
 
-    /* Get the thread's group */
-    group = (Object *)INST_DATA(jThread)[group_offset];
+    if(uncaught_mb != NULL) {
+        clearException();
+        executeMethod(handler, uncaught_mb, jThread, excep);
 
-    /* If there's an uncaught exception, call uncaughtException on the thread's
-       exception handler, or the thread's group if this is unset */
-    if((excep = exceptionOccurred())) {
-        FieldBlock *fb = findField(thread_class, SYMBOL(exceptionHandler),
-                                                 SYMBOL(sig_java_lang_Thread_UncaughtExceptionHandler));
-        Object *thread_handler = fb == NULL ? NULL : (Object *)INST_DATA(jThread)[fb->offset];
-        Object *handler = thread_handler == NULL ? group : thread_handler;
-
-        MethodBlock *uncaught_exp = lookupMethod(handler->class, SYMBOL(uncaughtException),
-                                                 SYMBOL(_java_lang_Thread_java_lang_Throwable__V));
-
-        if(uncaught_exp) {
-            clearException();
-            executeMethod(handler, uncaught_exp, jThread, excep);
-        } else
-            printException();
+        /* If an exception occurred while trying to handle
+           the exception reinstate the original exception. */
+        if(exceptionOccurred())
+            setException(excep);
     }
 
+    /* If no method is found, or an error occurred, try to print
+       the exception (if it was handled, no exception will be
+       pending and the call will do nothing) */
+    printException();
+}
+
+void detachThread(Thread *thread) {
+    ExecEnv *ee = thread->ee;
+    Object *jThread = ee->thread;
+    Object *group = INST_DATA(jThread, Object*, group_offset);
+    Object *vmthread = INST_DATA(jThread, Object*, vmthread_offset);
+
+    /* If there's an exception pending, it is uncaught */
+    if(exceptionOccurred0(ee))
+        uncaughtException();
+
     /* remove thread from thread group */
-    executeMethod(group, (CLASS_CB(group->class))->method_table[rmveThrd_mtbl_idx], jThread);
+    executeMethod(group, (CLASS_CB(group->class))->
+                                     method_table[rmveThrd_mtbl_idx], jThread);
 
     /* set VMThread ref in Thread object to null - operations after this
        point will result in an IllegalThreadStateException */
-    INST_DATA(jThread)[vmthread_offset] = 0;
+    INST_DATA(jThread, Object*, vmthread_offset) = NULL;
 
     /* Remove thread from the ID map hash table */
     deleteThreadFromHash(thread);
@@ -498,7 +643,7 @@ void detachThread(Thread *thread) {
     freeThreadID(thread->id);
 
     /* Handle daemon thread status */
-    if(!INST_DATA(jThread)[daemon_offset])
+    if(!INST_DATA(jThread, int, daemon_offset))
         non_daemon_thrds--;
 
     pthread_mutex_unlock(&lock);
@@ -524,12 +669,15 @@ void detachThread(Thread *thread) {
 
     if(non_daemon_thrds == 0) {
         /* No need to bother with disabling suspension
-         * around lock, as we're no longer on thread list */
+           around lock, as we're no longer on thread list */
         pthread_mutex_lock(&exit_lock);
         pthread_cond_signal(&exit_cv);
         pthread_mutex_unlock(&exit_lock);
     }
 
+    /* Finally, clear the thread local data */
+    setThreadSelf(NULL);
+
     TRACE("Thread 0x%x id: %d detached from VM\n", thread, thread->id);
 }
 
@@ -543,7 +691,7 @@ void *threadStart(void *arg) {
 
     /* Complete initialisation of the thread structure, create the thread
        stack and add the thread to the thread list */
-    initThread(thread, INST_DATA(jThread)[daemon_offset], &thread);
+    initThread(thread, INST_DATA(jThread, int, daemon_offset), &thread);
 
     /* Add thread to thread ID map hash table. */
     addThreadToHash(thread);
@@ -570,10 +718,11 @@ void createJavaThread(Object *jThread, long long stack_size) {
     disableSuspend(self);
 
     pthread_mutex_lock(&lock);
-    if(INST_DATA(jThread)[vmthread_offset]) {
+    if(INST_DATA(jThread, Object*, vmthread_offset) != NULL) {
         pthread_mutex_unlock(&lock);
         enableSuspend(self);
-        signalException(java_lang_IllegalThreadStateException, "thread already started");
+        signalException(java_lang_IllegalThreadStateException,
+                        "thread already started");
         return;
     }
 
@@ -586,14 +735,14 @@ void createJavaThread(Object *jThread, long long stack_size) {
     ee->thread = jThread;
     ee->stack_size = stack_size;
 
-    INST_DATA(vmthread)[vmData_offset] = (uintptr_t)thread;
-    INST_DATA(vmthread)[thread_offset] = (uintptr_t)jThread;
-    INST_DATA(jThread)[vmthread_offset] = (uintptr_t)vmthread;
+    INST_DATA(vmthread, Thread*, vmData_offset) = thread;
+    INST_DATA(vmthread, Object*, thread_offset) = jThread;
+    INST_DATA(jThread, Object*, vmthread_offset) = vmthread;
 
     pthread_mutex_unlock(&lock);
 
     if(pthread_create(&thread->tid, &attributes, threadStart, thread)) {
-        INST_DATA(jThread)[vmthread_offset] = 0;
+        INST_DATA(jThread, Object*, vmthread_offset) = NULL;
         sysFree(ee);
         enableSuspend(self);
         signalException(java_lang_OutOfMemoryError, "can't create thread");
@@ -618,7 +767,7 @@ Thread *attachJNIThread(char *name, char is_daemon, Object *group) {
 
     /* If no group is given add it to the main group */
     if(group == NULL)
-        group = (Object*)INST_DATA(main_ee.thread)[group_offset];
+        group = INST_DATA(main_ee.thread, Object*, group_offset);
 
     /* Initialise internal thread structure */
     memset(thread, 0, sizeof(Thread));
@@ -646,7 +795,7 @@ static void *shell(void *args) {
     /* VM helper threads should be added to the system group, but this doesn't
        exist.  As the root group is main, we add it to that for now... */
     attachThread(((char**)args)[0], TRUE, &self, self,
-                 (Object*)INST_DATA(main_ee.thread)[group_offset]);
+                 INST_DATA(main_ee.thread, Object*, group_offset));
 
     sysFree(args);
     (*(void(*)(Thread*))start)(self);
@@ -678,12 +827,14 @@ void suspendThread(Thread *thread) {
     MBARRIER();
 
     if(!thread->blocking) {
-        TRACE("Sending suspend signal to thread 0x%x id: %d\n", thread, thread->id);
+        TRACE("Sending suspend signal to thread 0x%x id: %d\n",
+              thread, thread->id);
         pthread_kill(thread->tid, SIGUSR1);
     }
 
     while(thread->blocking != SUSP_BLOCKING && thread->state != SUSPENDED) {
-        TRACE("Waiting for thread 0x%x id: %d to suspend\n", thread, thread->id);
+        TRACE("Waiting for thread 0x%x id: %d to suspend\n",
+              thread, thread->id);
         sched_yield();
     }
 }
@@ -693,7 +844,8 @@ void resumeThread(Thread *thread) {
     MBARRIER();
 
     if(!thread->blocking) {
-        TRACE("Sending resume signal to thread 0x%x id: %d\n", thread, thread->id);
+        TRACE("Sending resume signal to thread 0x%x id: %d\n",
+              thread, thread->id);
         pthread_kill(thread->tid, SIGUSR1);
     }
 
@@ -717,7 +869,8 @@ void suspendAllThreads(Thread *self) {
         MBARRIER();
 
         if(!thread->blocking) {
-            TRACE("Sending suspend signal to thread 0x%x id: %d\n", thread, thread->id);
+            TRACE("Sending suspend signal to thread 0x%x id: %d\n",
+                  thread, thread->id);
             pthread_kill(thread->tid, SIGUSR1);
         }
     }
@@ -727,7 +880,8 @@ void suspendAllThreads(Thread *self) {
             continue;
 
         while(thread->blocking != SUSP_BLOCKING && thread->state != SUSPENDED) {
-            TRACE("Waiting for thread 0x%x id: %d to suspend\n", thread, thread->id);
+            TRACE("Waiting for thread 0x%x id: %d to suspend\n",
+                  thread, thread->id);
             sched_yield();
         }
     }
@@ -752,14 +906,16 @@ void resumeAllThreads(Thread *self) {
         MBARRIER();
 
         if(!thread->blocking) {
-            TRACE("Sending resume signal to thread 0x%x id: %d\n", thread, thread->id);
+            TRACE("Sending resume signal to thread 0x%x id: %d\n",
+                  thread, thread->id);
             pthread_kill(thread->tid, SIGUSR1);
         }
     }
 
     for(thread = &main_thread; thread != NULL; thread = thread->next) {
         while(thread->state == SUSPENDED) {
-            TRACE("Waiting for thread 0x%x id: %d to resume\n", thread, thread->id);
+            TRACE("Waiting for thread 0x%x id: %d to resume\n",
+                  thread, thread->id);
             sched_yield();
         }
     }
@@ -767,7 +923,7 @@ void resumeAllThreads(Thread *self) {
     all_threads_suspended = FALSE;
     if(threads_waiting_to_start) {
         TRACE("%d threads waiting to start...\n", threads_waiting_to_start);
-	    pthread_cond_broadcast(&cv);
+        pthread_cond_broadcast(&cv);
     }
 
     TRACE("All threads resumed...\n");
@@ -851,8 +1007,11 @@ void fastEnableSuspend(Thread *thread) {
         sigaddset(&mask, SIGUSR1);
         pthread_sigmask(SIG_BLOCK, &mask, NULL);
 
-        TRACE("Thread 0x%x id: %d is fast self suspending\n", thread, thread->id);
+        TRACE("Thread 0x%x id: %d is fast self suspending\n",
+              thread, thread->id);
+
         suspendLoop(thread);
+
         TRACE("Thread 0x%x id: %d resumed\n", thread, thread->id);
 
         pthread_sigmask(SIG_UNBLOCK, &mask, NULL);
@@ -880,29 +1039,34 @@ void dumpThreadsLoop(Thread *self) {
         /* It must be a SIGQUIT.  Do a thread dump */
 
         suspendAllThreads(self);
-        jam_printf("\n------ JamVM version %s Full Thread Dump -------\n", VERSION);
+        jam_printf("\n------ JamVM version %s Full Thread Dump -------\n",
+                   VERSION);
 
         for(thread = &main_thread; thread != NULL; thread = thread->next) {
-            uintptr_t *thr_data = INST_DATA(thread->ee->thread);
-            int priority = thr_data[priority_offset];
-            int daemon = thr_data[daemon_offset];
+            Object *jThread = thread->ee->thread;
+            int priority = INST_DATA(jThread, int, priority_offset);
+            int daemon = INST_DATA(jThread, int, daemon_offset);
             Frame *last = thread->ee->last_frame;
 
-            /* Get thread name; we don't use String2Cstr(), as this mallocs memory
-               and may deadlock with a thread suspended in malloc/realloc/free */
-            String2Buff((Object*)thr_data[name_offset], buffer, sizeof(buffer));
+            /* Get thread name; we don't use String2Cstr(), as this mallocs
+               memory and may deadlock with a thread suspended in
+               malloc/realloc/free */
+            String2Buff(INST_DATA(jThread, Object*, name_offset), buffer,
+                        sizeof(buffer));
 
-            jam_printf("\n\"%s\"%s %p priority: %d tid: %p id: %d state: %s (%d)\n",
-                  buffer, daemon ? " (daemon)" : "", thread, priority, thread->tid,
-                  thread->id, getThreadStateString(thread), thread->state);
+            jam_printf("\n\"%s\"%s %p priority: %d tid: %p id: %d state:"
+                       "%s (%d)\n", buffer, daemon ? " (daemon)" : "",
+                       thread, priority, thread->tid, thread->id,
+                       getThreadStateString(thread), thread->state);
 
             while(last->prev != NULL) {
                 for(; last->mb != NULL; last = last->prev) {
                     MethodBlock *mb = last->mb;
                     ClassBlock *cb = CLASS_CB(mb->class);
 
-                    /* Convert slashes in class name to dots.  Similar to above,
-                       we don't use slash2dots(), as this mallocs memory */
+                    /* Convert slashes in class name to dots.  Similar to
+                       above, we don't use slash2dots(), as this mallocs
+                       memory */
                     slash2dots2buff(cb->name, buffer, sizeof(buffer)); 
                     jam_printf("\tat %s.%s(", buffer, mb->name);
 
@@ -1003,7 +1167,7 @@ void mainThreadSetContextClassLoader(Object *loader) {
     FieldBlock *fb = findField(thread_class, SYMBOL(contextClassLoader),
                                              SYMBOL(sig_java_lang_ClassLoader));
     if(fb != NULL)
-        INST_DATA(main_ee.thread)[fb->offset] = (uintptr_t)loader;
+        INST_DATA(main_ee.thread, Object*, fb->u.offset) = loader;
 }
 
 void initialiseThreadStage1(InitArgs *args) {
@@ -1013,14 +1177,16 @@ void initialiseThreadStage1(InitArgs *args) {
     dflt_stack_size = args->java_stack;
 
     /* Initialise internal locks and pthread state */
-    pthread_key_create(&threadKey, NULL);
-
     pthread_mutex_init(&lock, NULL);
     pthread_cond_init(&cv, NULL);
 
     pthread_mutex_init(&exit_lock, NULL);
     pthread_cond_init(&exit_cv, NULL);
 
+#ifndef HAVE_TLS
+    pthread_key_create(&self, NULL);
+#endif
+
     pthread_attr_init(&attributes);
     pthread_attr_setdetachstate(&attributes, PTHREAD_CREATE_DETACHED);
 
@@ -1046,74 +1212,89 @@ void initialiseThreadStage1(InitArgs *args) {
 
     initialiseJavaStack(&main_ee);
     setThreadSelf(&main_thread);
+
+    pthread_cond_init(&main_thread.wait_cv, NULL);
+
+    main_thread.park_state = PARK_RUNNING;
+    pthread_cond_init(&main_thread.park_cv, NULL);
+    pthread_mutex_init(&main_thread.park_lock, NULL);
 }
 
 void initialiseThreadStage2(InitArgs *args) {
     Object *java_thread;
     Class *thrdGrp_class;
     MethodBlock *run, *remove_thread;
-    FieldBlock *group, *priority, *root, *threadId;
-    FieldBlock *vmThread = NULL, *thread = NULL;
     FieldBlock *vmData, *daemon, *name;
+    FieldBlock *vmThread, *thread, *group;
+    FieldBlock *priority, *root, *threadId;
 
     /* Load thread class and register reference for compaction threading */
-    thread_class = findSystemClass0(SYMBOL(java_lang_Thread));
-    registerStaticClassRef(&thread_class);
-
-    if(thread_class != NULL) {
-        vmThread = findField(thread_class, SYMBOL(vmThread), SYMBOL(sig_java_lang_VMThread));
-        daemon = findField(thread_class, SYMBOL(daemon), SYMBOL(Z));
-        name = findField(thread_class, SYMBOL(name), SYMBOL(sig_java_lang_String));
-        group = findField(thread_class, SYMBOL(group), SYMBOL(sig_java_lang_ThreadGroup));
-        priority = findField(thread_class, SYMBOL(priority), SYMBOL(I));
-        threadId = findField(thread_class, SYMBOL(threadId), SYMBOL(J));
-
-        init_mb = findMethod(thread_class, SYMBOL(object_init),
-                             SYMBOL(_java_lang_VMThread_java_lang_String_I_Z__V));
-        run = findMethod(thread_class, SYMBOL(run), SYMBOL(___V));
-
-        vmthread_class = findSystemClass0(SYMBOL(java_lang_VMThread));
-        CLASS_CB(vmthread_class)->flags |= VMTHREAD;
-
-        /* Register class reference for compaction threading */
-        registerStaticClassRef(&vmthread_class);
-
-        if(vmthread_class != NULL) {
-            thread = findField(vmthread_class, SYMBOL(thread), SYMBOL(sig_java_lang_Thread));
-            vmData = findField(vmthread_class, SYMBOL(vmData), SYMBOL(I));
-        }
-    }
-
-    /* findField and findMethod do not throw an exception... */
-    if((init_mb == NULL) || (vmData == NULL) || (run == NULL) || (daemon == NULL) ||
-       (name == NULL) || (group == NULL) || (priority == NULL) || (vmThread == NULL) ||
-       (thread == NULL) || (threadId == NULL))
+    if((thread_class = findSystemClass0(SYMBOL(java_lang_Thread))) == NULL)
         goto error;
 
-    vmthread_offset = vmThread->offset;
-    thread_offset = thread->offset;
-    vmData_offset = vmData->offset;
-    daemon_offset = daemon->offset;
-    group_offset = group->offset;
-    priority_offset = priority->offset;
-    threadId_offset = threadId->offset;
-    name_offset = name->offset;
+    registerStaticClassRef(&thread_class);
+
+    vmThread = findField(thread_class, SYMBOL(vmThread),
+                                       SYMBOL(sig_java_lang_VMThread));
+    daemon = findField(thread_class, SYMBOL(daemon), SYMBOL(Z));
+    name = findField(thread_class, SYMBOL(name), SYMBOL(sig_java_lang_String));
+    group = findField(thread_class, SYMBOL(group),
+                                    SYMBOL(sig_java_lang_ThreadGroup));
+    priority = findField(thread_class, SYMBOL(priority), SYMBOL(I));
+    threadId = findField(thread_class, SYMBOL(threadId), SYMBOL(J));
+
+    init_mb = findMethod(thread_class, SYMBOL(object_init),
+                         SYMBOL(_java_lang_VMThread_java_lang_String_I_Z__V));
+    run = findMethod(thread_class, SYMBOL(run), SYMBOL(___V));
+
+    if((vmthread_class = findSystemClass0(SYMBOL(java_lang_VMThread))) == NULL)
+        goto error;
+
+    CLASS_CB(vmthread_class)->flags |= VMTHREAD;
+
+    /* Register class reference for compaction threading */
+    registerStaticClassRef(&vmthread_class);
+
+    thread = findField(vmthread_class, SYMBOL(thread),
+                                       SYMBOL(sig_java_lang_Thread));
+    vmData = findField(vmthread_class, SYMBOL(vmData), SYMBOL(J));
+
+    /* findField and findMethod do not throw an exception... */
+    if((init_mb == NULL)  || (vmData == NULL)   || (run == NULL) ||
+       (daemon == NULL)   || (name == NULL)     || (group == NULL) ||
+       (priority == NULL) || (vmThread == NULL) || (thread == NULL) ||
+       (threadId == NULL))
+        goto error;
+
+    vmthread_offset = vmThread->u.offset;
+    thread_offset = thread->u.offset;
+    vmData_offset = vmData->u.offset;
+    daemon_offset = daemon->u.offset;
+    group_offset = group->u.offset;
+    priority_offset = priority->u.offset;
+    threadId_offset = threadId->u.offset;
+    name_offset = name->u.offset;
     run_mtbl_idx = run->method_table_index;
 
     /* Initialise the Java-level thread objects for the main thread */
-    java_thread = initJavaThread(&main_thread, FALSE, "main");
+    if((java_thread = initJavaThread(&main_thread, FALSE, "main")) == NULL)
+        goto error;
 
     /* Main thread is now sufficiently setup to be able to run the thread group
        initialiser.  This is essential to create the root thread group */
     thrdGrp_class = findSystemClass(SYMBOL(java_lang_ThreadGroup));
 
-    root = findField(thrdGrp_class, SYMBOL(root), SYMBOL(sig_java_lang_ThreadGroup));
+    if(exceptionOccurred())
+        goto error;
+
+    root = findField(thrdGrp_class, SYMBOL(root),
+                                    SYMBOL(sig_java_lang_ThreadGroup));
 
     addThread_mb = findMethod(thrdGrp_class, SYMBOL(addThread),
-                                             SYMBOL(_java_lang_Thread_args__void));
+                                             SYMBOL(_java_lang_Thread__V));
 
     remove_thread = findMethod(thrdGrp_class, SYMBOL(removeThread),
-                                              SYMBOL(_java_lang_Thread_args__void));
+                                              SYMBOL(_java_lang_Thread__V));
 
     /* findField and findMethod do not throw an exception... */
     if((root == NULL) || (addThread_mb == NULL) || (remove_thread == NULL))
@@ -1122,8 +1303,11 @@ void initialiseThreadStage2(InitArgs *args) {
     rmveThrd_mtbl_idx = remove_thread->method_table_index;
 
     /* Add the main thread to the root thread group */
-    INST_DATA(java_thread)[group_offset] = root->static_value;
-    executeMethod(((Object*)root->static_value), addThread_mb, java_thread);
+    INST_DATA(java_thread, Object*, group_offset) = root->u.static_value.p;
+    executeMethod(((Object*)root->u.static_value.p), addThread_mb, java_thread);
+
+    if(exceptionOccurred())
+        goto error;
 
     /* Setup signal handling.  This will be inherited by all
        threads created within Java */
