@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2003, 2004, 2005, 2006, 2007
+ * Copyright (C) 2003, 2004, 2005, 2006, 2007, 2009
  * Robert Lougher <rob@lougher.org.uk>.
  *
  * This file is part of JamVM.
@@ -29,17 +29,20 @@
 #include <zlib.h>
 #include <sys/mman.h>
 
+/* Required on OpenSolaris. */
+#ifndef MAP_FILE
+#define MAP_FILE 0
+#endif
+
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <fcntl.h>
 
-/* Definitions needed for hashtable */
+/* Common definitions needed for hashtable */
 #define HASHTABSZE 1<<8
-#define HASH(ptr) utf8Hash(ptr)
-#define COMPARE(ptr1, ptr2, hash1, hash2) (hash1 == hash2) && utf8Comp(ptr1, ptr2) 
 #define PREPARE(ptr) ptr
 #define SCAVENGE(ptr) FALSE
-#define FOUND(ptr) ptr
+#define FOUND(ptr1, ptr2) ptr2
 
 /* zlib window size */
 #define MAX_WINDOW_BITS 15
@@ -78,11 +81,49 @@
 #define COMP_STORED                0
 #define COMP_DEFLATED              8
 
+#define HASH(ptr) zipHash(ptr)
+#define COMPARE(ptr1, ptr2, hash1, hash2) ((hash1 == hash2) && \
+                                           zipComp(ptr1, ptr2)) 
+
+/* The filenames within the zip file are added to a hash-table for faster
+   access.  To save memory, the entries point directly into the zip file's
+   central directory.  However, filenames within the directory are not null-
+   terminated.  This means the path length must be read each time.  As the
+   utf8 routines operate on null-terminated strings, the filename must also
+   be copied into a temporary buffer.  This is NOT performance critical,
+   as the hash is calculated only once per entry (on insertion)
+*/
+int zipHash(char *path) {
+    int path_len = READ_LE_SHORT((unsigned char*)path +
+                                        (CEN_FILE_PATHLEN_OFFSET -
+                                         CEN_FILE_HEADER_LEN));
+    char buff[path_len + 1];
+
+    memcpy(buff, path, path_len);
+    buff[path_len] = '\0';
+
+    return utf8Hash(buff);
+}
+
+/* When adding entries to the hash-table, the paths being compared will
+   both be within the zip file.  As there are no concerns about differing
+   encoding of null, we can do a memcmp to compare them
+*/
+int zipComp(char *path1, char *path2) {
+    int path1_len = READ_LE_SHORT((unsigned char*)path1 +
+                                          (CEN_FILE_PATHLEN_OFFSET -
+                                           CEN_FILE_HEADER_LEN));
+    int path2_len = READ_LE_SHORT((unsigned char*)path2 +
+                                          (CEN_FILE_PATHLEN_OFFSET -
+                                           CEN_FILE_HEADER_LEN));
+
+    return path1_len == path2_len && !memcmp(path1, path2, path1_len);
+}
+
 ZipFile *processArchive(char *path) {
     unsigned char magic[SIG_LEN];
     unsigned char *data, *pntr;
     int entries, fd, len;
-    int next_file_sig;
 
     HashTable *hash_table;
     ZipFile *zip;
@@ -91,15 +132,15 @@ ZipFile *processArchive(char *path) {
         return NULL;
 
     /* First 4 bytes must be the signature for the first local file header */
-    if(read(fd, &magic[0], SIG_LEN) != SIG_LEN
-                     || READ_LE_INT(magic) != LOC_FILE_HEADER_SIG)
+    if(read(fd, &magic[0], SIG_LEN) != SIG_LEN ||
+                    READ_LE_INT(magic) != LOC_FILE_HEADER_SIG)
         goto error;
 
     /* Get the length */
     len = lseek(fd, 0, SEEK_END);
 
     /* Mmap the file into memory */
-    if((data = (unsigned char*)mmap(0, len, PROT_READ | PROT_WRITE, MAP_FILE |
+    if((data = (unsigned char*)mmap(0, len, PROT_READ, MAP_FILE |
                                             MAP_PRIVATE, fd, 0)) == MAP_FAILED)
         goto error;
 
@@ -129,7 +170,7 @@ ZipFile *processArchive(char *path) {
        Do not create lock -- we're single threaded (bootstrapping)
        and once entries are added, table is read-only */
 
-    hash_table = (HashTable*)sysMalloc(sizeof(HashTable));
+    hash_table = sysMalloc(sizeof(HashTable));
     initHashTable((*hash_table), HASHTABSZE, FALSE);
 
     /* Get the offset from the start of the file of the first directory entry */
@@ -137,25 +178,16 @@ ZipFile *processArchive(char *path) {
 
     /* Scan the directory list and add the entries to the hash table */
 
-    /* We terminate each pathname "in place" to save memory -- this may
-       over-write the signature for the next entry.  We therefore
-       speculatively load the next sig *before* we terminate the pathname */
-
-    if((pntr + SIG_LEN) > (data + len))
-        goto error2;
-
-    /* Speculatively read first sig */
-    next_file_sig = READ_LE_INT(pntr);
-
     while(entries--) {
-        char *found, *pathname, *term;
+        char *found, *pathname;
         int path_len, comment_len, extra_len;
 
+        /* Make sure we're not reading outside the file */
         if((pntr + CEN_FILE_HEADER_LEN) > (data + len))
             goto error2;
 
         /* Check directory entry signature is present */
-        if(next_file_sig != CEN_FILE_HEADER_SIG)
+        if(READ_LE_INT(pntr) != CEN_FILE_HEADER_SIG)
             goto error2;
 
         /* Get the length of the pathname */
@@ -167,22 +199,9 @@ ZipFile *processArchive(char *path) {
 
         /* The pathname starts after the fixed part of the dir entry */
         pathname = (char*)(pntr += CEN_FILE_HEADER_LEN);
-        term = (char*)(pntr += path_len);
 
-        /* Skip rest of variable fields -- should then be pointing to next
-           sig.  If no extra or comment fields, pntr == term */
-        pntr += extra_len + comment_len;
-
-        /* Even if this is the last entry, a well-formed zip file will *always*
-           have at least 4 extra bytes */
-        if((pntr + SIG_LEN) > (data + len))
-            goto error2;
-
-        /* Speculatively read next sig */
-        next_file_sig = READ_LE_INT(pntr);
-
-        /* Terminate the pathname */    
-        *term = '\0';
+        /* Skip variable fields, to point to next sig */
+        pntr += path_len + extra_len + comment_len;
 
         /* Add if absent, no scavenge, not locked */
         findHashEntry((*hash_table), pathname, found, TRUE, FALSE, FALSE);
@@ -204,6 +223,31 @@ error:
     return NULL;
 }
 
+#undef HASH
+#undef COMPARE
+
+#define HASH(ptr) utf8Hash(ptr)
+#define COMPARE(ptr1, ptr2, hash1, hash2) ((hash1 == hash2) && \
+                                           utf8ZipComp(ptr1, ptr2)) 
+
+/* When searching for an entry, the comparison will be between a null-
+   terminated utf8 path and a non null-terminated path within the zip.
+   Again, the memcpy is not performance critical, as comparisons are
+   performed using the hash value; a filename comparison is only
+   done when hash values clash (which is rare)
+*/
+int utf8ZipComp(char *path1, char *path2) {
+    int path2_len = READ_LE_SHORT((unsigned char*)path2 +
+                                          (CEN_FILE_PATHLEN_OFFSET -
+                                           CEN_FILE_HEADER_LEN));
+    char buff[path2_len + 1];
+
+    memcpy(buff, path2, path2_len);
+    buff[path2_len] = '\0';
+
+    return utf8Comp(path1, buff);
+}
+
 char *findArchiveDirEntry(char *pathname, ZipFile *zip) {
     char *found;
 
@@ -218,7 +262,8 @@ char *findArchiveEntry(char *pathname, ZipFile *zip, int *uncomp_len) {
     unsigned char *decomp_buff, *comp_data;
     unsigned char *dir_entry;
 
-    if((dir_entry = (unsigned char *)findArchiveDirEntry(pathname, zip)) == NULL)
+    dir_entry = (unsigned char *)findArchiveDirEntry(pathname, zip);
+    if(dir_entry == NULL)
         return NULL;
 
     /* Found the file -- the pathname points directly into the
@@ -234,17 +279,21 @@ char *findArchiveEntry(char *pathname, ZipFile *zip, int *uncomp_len) {
     /* Get the variable length part of the local file header */
     extra_len = READ_LE_SHORT(zip->data + offset + LOC_FILE_EXTRA_OFFSET);
 
-    /* Get the path_len again -- faster than doing a strlen */
-    path_len = READ_LE_SHORT(dir_entry + (CEN_FILE_PATHLEN_OFFSET - CEN_FILE_HEADER_LEN));
+    /* Get the path_len again */
+    path_len = READ_LE_SHORT(dir_entry + (CEN_FILE_PATHLEN_OFFSET -
+                                          CEN_FILE_HEADER_LEN));
 
     /* The file's length when uncompressed -- this is passed out */
-    *uncomp_len = READ_LE_INT(dir_entry + (CEN_FILE_UNCOMPLEN_OFFSET - CEN_FILE_HEADER_LEN));
+    *uncomp_len = READ_LE_INT(dir_entry + (CEN_FILE_UNCOMPLEN_OFFSET -
+                                           CEN_FILE_HEADER_LEN));
 
     /* The compressed file's length, i.e. the data size in the file */
-    comp_len = READ_LE_INT(dir_entry + (CEN_FILE_COMPLEN_OFFSET - CEN_FILE_HEADER_LEN));
+    comp_len = READ_LE_INT(dir_entry + (CEN_FILE_COMPLEN_OFFSET -
+                                        CEN_FILE_HEADER_LEN));
 
     /* How the file is compressed */
-    comp_method = READ_LE_SHORT(dir_entry + (CEN_FILE_COMPMETH_OFFSET - CEN_FILE_HEADER_LEN));
+    comp_method = READ_LE_SHORT(dir_entry + (CEN_FILE_COMPMETH_OFFSET -
+                                             CEN_FILE_HEADER_LEN));
 
     /* Calculate the data start */
     offset += LOC_FILE_HEADER_LEN + path_len + extra_len;
